#ifndef YSADVGEOMETRY_IS_INCLUDED
#define YSADVGEOMETRY_IS_INCLUDED
/* { */

YSBOOL YsCheckConvex2(int np,const YsVec2 p[],YSBOOL strictCheck=YSFALSE);
YSBOOL YsCheckConvex3(int np,const YsVec3 p[],YSBOOL strictCheck=YSFALSE);
YSBOOL YsCheckConvexByAngle2(int np,const YsVec2 p[]);
YSBOOL YsCheckConvexByAngle3(int np,const YsVec3 p[]);
double YsGetTriangleArea3(const YsVec3 &p1,const YsVec3 &p2,const YsVec3 &p3);
double YsGetTriangleArea2(const YsVec2 &p1,const YsVec2 &p2,const YsVec2 &p3);
YSRESULT YsGetLargestTriangleFromPolygon3
   (const YsVec3 *v[3],int np,const YsVec3 p[]);
YSRESULT YsGetLargestTriangleFromPolygon3(YsVec3 v[3],int np,const YsVec3 p[]);
YSRESULT YsGetLargestTriangleFromPolygon2
   (const YsVec2 *v[3],int np,const YsVec2 p[]);
YSRESULT YsGetLargestTriangleFromPolygon2(YsVec2 v[3],int np,const YsVec2 p[]);
YSRESULT YsGetAverageNormalVector(YsVec3 &nom,int np,const YsVec3 p[]);
YSBOOL YsCheckInBetween2(const YsVec2 &q,const YsVec2 &p1,const YsVec2 &p2);
YSBOOL YsCheckInBetween3(const YsVec3 &q,const YsVec3 &p1,const YsVec3 &p2);
YSRESULT YsGetLineIntersection2
   (YsVec2 &crs,const YsVec2 &p1,const YsVec2 &p2,const YsVec2 &q1,const YsVec2 &q2);
YSBOOL YsCheckLineOverlap2
    (const YsVec2 &o1,const YsVec2 &v1,const YsVec2 &o2,const YsVec2 &v2);
YSBOOL YsCheckLineOverlap3
    (const YsVec3 &o1,const YsVec3 &v1,const YsVec3 &o2,const YsVec3 &v2);
YSSIDE YsCheckInsidePolygon2(const YsVec2 &ref,int np,const YsVec2 p[]);
YSSIDE YsCheckInsidePolygon3(const YsVec3 &ref,int np,const YsVec3 p[]);

YSRESULT YsGetCenterOfPolygon3(YsVec3 &cen,int np,const YsVec3 plg[]);
YSRESULT YsGetCenterOfPolygon2(YsVec2 &cen,int np,const YsVec2 plg[]);
	// Note:These functions will return the center of the
	//      largest triangle given by YsGetLargestTriangleFromPolygon3/2

YSRESULT YsGetPolygonProjectionMatrix(YsMatrix4x4 &mat,int np,const YsVec3 p[]);
	// Returns a matrix that project a Polygon into XY plane.
	// Note:Twist direction will be reversed in Left-Handed coordinate.

YSINTERSECTION YsGetLinePenetration2
    (const YsVec2 &p1,const YsVec2 &p2,const YsVec2 &p3,const YsVec2 &p4);
YSINTERSECTION YsGetLinePenetration3
    (const YsVec3 &p1,const YsVec3 &p2,const YsVec3 &q1,const YsVec3 &q2);


YSBOOL YsCheckSeparatability2(int np,const YsVec2 p[],int idx1,int idx2);
YSBOOL YsCheckSeparatability3(int np,const YsVec3 p[],int idx1,int idx2);
	// Note:YsCheckSeparatability3 is extremely slow function.
	//      Not recommended to used it in a loop.
	//      It is recommended to make PolygonProjectionMatrix
	//      and project the polygon into XY plane. Then, use it to
	//      Check separatability.

YSBOOL YsCheckBoundingBoxCollision3
 (const YsVec3 &min1,const YsVec3 &max1,const YsVec3 &min2,const YsVec3 &max2);
YSBOOL YsCheckBoundingBoxCollision2
 (const YsVec2 &min1,const YsVec2 &max1,const YsVec2 &min2,const YsVec2 &max2);

YSBOOL YsCheckInsideBoundingBox3
    (const YsVec3 &ref,const YsVec3 &min,const YsVec3 &max);
YSBOOL YsCheckInsideBoundingBox2
    (const YsVec2 &ref,const YsVec2 &min,const YsVec2 &max);

YSFLIPDIRECTION YsCheckFlipDirection2(int np,const YsVec2 p[]);
YSFLIPDIRECTION YsCheckFlipDirection3(int np,const YsVec3 p[],const YsVec3 &n);

YSRESULT YsGetTwoPlaneCrossLine
    (YsVec3 &lnOrg,YsVec3 &lnVec,const YsPlane &pl1,const YsPlane &pl2);

YSRESULT YsGetNearestPointOfTwoLine
    (YsVec3 &np,YsVec3 &nq,
     const YsVec3 &p1,const YsVec3 &p2,const YsVec3 &q1,const YsVec3 &q2);
YSRESULT YsGetNearestPointOnLine3
    (YsVec3 &np,const YsVec3 &p1,const YsVec3 &p2,const YsVec3 &ref);
YSRESULT YsGetNearestPointOnLine2
    (YsVec2 &np,const YsVec2 &p1,const YsVec2 &p2,const YsVec2 &ref);


YSINTERSECTION YsGetPolygonPenetration
    (int np,const YsVec3 p[],int nq,const YsVec3 q[]);

YSRESULT YsGetArbitraryInsidePointOfPolygon2
  (YsVec2 &aip,int np,const YsVec3 p[],YSBOOL strictCheckOfConvexity=YSFALSE);
YSRESULT YsGetArbitraryInsidePointOfPolygon3
  (YsVec3 &aip,int np,const YsVec3 p[],YSBOOL strictCheckOfConvexity=YSFALSE);

/* } */
#endif
