#ifndef YSARRAY_IS_INCLUDED
#define YSARRAY_IS_INCLUDED
/* { */

template <class T> class YsArray
{
// CAUTION!
// Whenever add an item, do not forget to modify operator=

public:
	YsArray();
	YsArray(int nv,const T v[]);
	YsArray(const YsArray <T> &from);
	virtual ~YsArray();

	const YsArray <T> &operator=(const YsArray <T> &from);
	const YsArray <T> &operator=(const YsList <T> *from);

	enum
	{
		UnitLength=16
	};

	YSRESULT Set(int nv,const T v[]);
	inline YSRESULT SetItem(int idx,const T &v)
	{
		if(0<=idx && idx<nv)
		{
			vv[idx]=v;
			return YSOK;
		}
		return YSERR;
	}
	inline int GetNumItem(void) const
	{
		return nv;
	}
	inline const T &v(int i) const
	{
		if(0<=i && i<nv)
		{
			return vv[i];
		}
		else
		{
			YsPrintf::def->Printf
			    ("YsArray::v()\nArray Subscript Exceeded\n");
			return *((T *)NULL);
		}
	}
	inline const T &GetItem(int i) const
	{
		return v(i);
	}
	inline const T *GetArray(void) const
	{
		return vv;
	}

	YSRESULT AppendItem(const T &dat);

protected:
	int nv;
	T *vv;
	inline YSRESULT Alloc(int n,YSBOOL cpy);
};

template <class T> inline YSRESULT YsArray<T>::Alloc(int n,YSBOOL cpy)
{
	T *neo;
	int allocSize,currentSize;
	allocSize=(n+UnitLength-1)/UnitLength*UnitLength;
	currentSize=(nv+UnitLength-1)/UnitLength*UnitLength;
	if(allocSize<=currentSize)
	{
		nv=n;
		return YSOK;
	}
	else
	{
		neo=new T [allocSize];
		if(neo!=NULL)
		{
			if(cpy==YSTRUE)
			{
				int i;
				for(i=0; i<nv; i++)
				{
					neo[i]=vv[i];
				}
			}
			if(vv!=NULL)
			{
				delete [] vv;
			}
			nv=n;
			vv=neo;
			return YSOK;
		}
		else
		{
			return YSERR;
		}
	}
}

template <class T> YsArray<T>::YsArray()
{
	nv=0;
	vv=NULL;
}

template <class T> YsArray<T>::YsArray(int nv,const T v[])
{
	nv=0;
	vv=NULL;
	Set(nv,vv);
}

template <class T> YsArray<T>::YsArray(const YsArray <T> &from)
{
	nv=0;
	vv=NULL;
	*this=from;
}

template <class T> YsArray<T>::~YsArray()
{
	if(vv!=0)
	{
		delete [] vv;
	}
}

template <class T>
const YsArray<T> &YsArray<T>::operator=(const YsArray <T> &from)
{
	if(this!=&from)
	{
		Set(from.nv,from.vv);
	}
	return *this;
}

template <class T> YSRESULT YsArray<T>::Set(int NV,const T V[])
{
	if(NV==0)
	{
		if(vv!=NULL)
		{
			delete [] vv;
		}
		nv=0;
		vv=NULL;
		return YSOK;
	}
	else
	{
		if(Alloc(NV,YSFALSE)==YSOK)
		{
			for(int i=0; i<NV; i++)
			{
				vv[i]=V[i];
			}
			return YSOK;
		}
		else
		{
			YsPrintf::def->Printf("YsArray::Set()\nLow Memory Warning!!\n");
			return YSERR;
		}
	}
}

template <class T>
const YsArray<T> &YsArray<T>::operator=(const YsList <T> *from)
{
	int nItem;
	const YsList <T> *seeker;

	nItem=from->GetNumObject();
	if(Alloc(nItem,YSFALSE)==YSOK)
	{
		int i;
		for(i=0; i<nItem; i++)
		{
			seeker=from->Seek(i);
			vv[i]=seeker->dat;
		}
	}
	else
	{
		YsPrintf::def->Printf("YsArray::Set()\nLow Memory Warning!!\n");
	}
	return *this;
}

template <class T> YSRESULT YsArray<T>::AppendItem(const T &dat)
{
	if(Alloc(nv+1,YSTRUE)==YSOK)
	{
		vv[nv-1]=dat;
		return YSOK;
	}
	return YSERR;
}

/* } */
#endif
