#ifndef YSSHELL_IS_INCLUDED
#define YSSHELL_IS_INCLUDED
/* { */

#ifdef YS_COMPILING_YSSHELL_CPP
class YsShell;
static YSRESULT YsBlendShell1to2
    (YsShell &newShell,
     const YsShell &sh1,const YsShell &sh2,int f,YSBOOL elimOverlap);
#endif


// YsBaseClass
//   YsShellItem
//     YsShellVertex
//     YsShellPolygon
//   YsShell

// Declaration /////////////////////////////////////////////
class YsShellItem : public YsBaseClass
{
public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsShellItem::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsShellItem::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsShellItem();
	YsShellItem(const YsShellItem &from);
	virtual ~YsShellItem();

	int freeAttribute0; // Use free for any purpose
	int freeAttribute1; // Use free for any purpose
	int freeAttribute2; // Use free for any purpose
	int freeAttribute3; // Use free for any purpose
	int freeAttribute4;
};

// Declaration /////////////////////////////////////////////
class YsShellVertex : public YsShellItem
{
public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsShellVertex::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsShellVertex::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsShellVertex();
	YsShellVertex(const YsShellVertex &from);
	virtual ~YsShellVertex();

	YSRESULT AddUsedBy(int plId);
	YSRESULT DeleteUsedBy(int plId);
	YSRESULT SetPosition(const YsVec3 &pos);

	YSRESULT AfterDeletePolygon(int deletedPlId);

protected:
	const YsVec3 &GetPosition(void) const;
	YsVec3 pos,nom;
	YsArray <int> usedBy;

	friend class YsShell;
};

// Declaration /////////////////////////////////////////////
class YsShellPolygon : public YsShellItem
{
public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsShellPolygon::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsShellPolygon::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsShellPolygon();
	YsShellPolygon(const YsShellPolygon &from);
	virtual ~YsShellPolygon();

protected:
	YsVec3 nom;
	YsColor col;
	YsArray <int> idx;
	void Invert(void);

	friend class YsShell;
	friend YSRESULT YsBlendShell1to2
	    (YsShell &,const YsShell &,const YsShell &,int,YSBOOL);
};

// Declaration /////////////////////////////////////////////
class YsShell : public YsBaseClass
{
// CAUTION!
// Whenever add an item, do not forget to modify operator=

public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsShell::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsShell::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsShell();
	YsShell(const YsShell &from);
	virtual ~YsShell();

	void CleanUp(void);

	void Encache(void) const;

	int AddVertex(const YsVec3 &vec);
	int AddPolygon(int nv,const int v[]);
	int AddPolygonVertex(int nv,const YsVec3 v[]);

	YSRESULT ModifyVertexPosition(int vtId,const YsVec3 &neoPos);

	YSRESULT ModifyPolygon
	         (int plId,int nv,const int v[],YSBOOL autoUpdateUsedBy=YSTRUE);
	YSRESULT ApplyVtIdConversionTable
	         (int nEntry,const int tab[],YSBOOL autoModifyUsedBy=YSTRUE);

	YSRESULT GetNormalOfPolygon(YsVec3 &nom,int plId) const;
	YSRESULT GetColorOfPolygon(YsColor &col,int plId) const;

	YSRESULT SetNormalOfPolygon(int plId,const YsVec3 &nom);
	YSRESULT SetColorOfPolygon(int plId,const YsColor &col);

	YSRESULT ValidateVtId(int nVtId,const int vtId[]) const;
	YSRESULT ValidatePlId(int nPlId,const int plId[]) const;

	YSRESULT DeletePolygon(int plId,YSBOOL autoUpdateUsedBy=YSTRUE);
	YSRESULT DeleteVertex(int vtId,YSBOOL autoUpdateVtIdOfPolygon=YSTRUE);
	  // Note:if autoUpdateVtIdOfPolygon is YSTRUE, DeleteVertex
	  //      will not check if the vertex is used or not. Use caution.
	YSRESULT DeleteVertexAtTheSamePosition(void);
	YSRESULT DeleteVertexAtTheSamePosition(int &nDeleted);

	const YsShell &operator=(const YsShell &from);

	int GetNumPolygon(void) const;
	int GetNumVertex(void) const;
	int GetMaxNumVertexOfPolygon(void) const;

	YSRESULT GetVertexPosition(YsVec3 &pos,int vtId) const;

	void GetBoundingBox(YsVec3 &min,YsVec3 &max);

	int GetNumVertexOfPolygon(int plId) const;
	YSRESULT GetVertexListOfPolygon(int idx[],int maxcount,int plId) const;
	YSRESULT GetVertexListOfPolygon(YsVec3 vtx[],int maxcount,int plId) const;
	const int *GetVertexListOfPolygon(int plId) const;

	void SetMatrix(const YsMatrix4x4 &mat);
	void GetMatrix(YsMatrix4x4 &mat) const;

	YsShellPolygon *GetPolygon(int id);
	YsShellVertex *GetVertex(int id);
	const YsShellPolygon *GetPolygon(int id) const;
	const YsShellVertex *GetVertex(int id) const;

	YSSIDE CountRayIntersection
	    (int &CrossCounter,const YsVec3 &org,const YsVec3 &vec,int plIdSkip=-1)
	     const;

	YSSIDE CheckInsideSolid(const YsVec3 &pnt) const;

	YSRESULT InvertPolygon(int plId);

	int ShootRay(YsVec3 &intersect,const YsVec3 &p1,const YsVec3 &vec) const;
	    // Return PlId
	void ShootRay(YsArray <YsVec3> &intersect,const YsVec3 &p1,const YsVec3 &vec) const;
	    // Return PlId

	YSRESULT OmitPolygonAfterBlend(void);

	YSRESULT RefreshVertexReferenceTable(void);

// Members to copy in operator=
protected:
	YsMatrix4x4 mat;
	YsList <YsShellPolygon> *plg;
	YsList <YsShellVertex> *vtx;



	int AddPolygonVertex(int nv,const YsVec3 v[],int fa3,const int fa4[]);

	YSRESULT BlendPolygonVtId
	    (int &nvNew,int vtIdNew[],int nBuf,
	     int nvOrg,const int vtIdOrg[],int nv,const int vtId[]);
	YSBOOL AdjustBlendedPolygon(int &nv,int v[]);

// Isolated Stuff
public:
	int srfReadMode,srfReadPlId;
	YsColor srfReadColor;
	YsVec3 srfReadNom;
	YSRESULT BeginReadSrf(void);
	YSRESULT ReadSrfOneLine(char str[]);
	YSRESULT EndReadSrf(void);

	YSRESULT LoadSrf(char fn[]);

	friend YSRESULT YsBlendShell1to2
	    (YsShell &,const YsShell &,const YsShell &,int,YSBOOL);
};


YSBOOL YsCheckShellCollision(YsShell &sh1,YsShell &sh2);
	// Note:This function will use freeAttribute3 of YsShellPolygon

YSRESULT YsBlendShell
    (YsShell &newShell,
     const YsShell &sh1,
     const YsShell &sh2,
     YSBOOLEANOPERATION boolop=YSBOOLBLEND);
    // Note:This function will use freeAttribute2 and 3.

YSRESULT YsSeparateShell
    (YsShell &sh1,YsShell &sh2,const YsShell &org,const YsShell &ref);
    // Note:This function will use freeAttribute2,3 and 4.
    // sh1=org-ref
    // sh2=org&ref
/* } */
#endif
