#ifndef YSWIZARD_IS_INCLUDED
#define YSWIZARD_IS_INCLUDED
/* { */

// Declaration /////////////////////////////////////////////
class YsWizard : public YsBaseClass
{
public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsWizard::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsWizard::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsWizard();
	YsWizard(const YsWizard &from);
	virtual ~YsWizard();
};

// Declaration /////////////////////////////////////////////
class YsCollisionOfPolygon : public YsWizard
{
public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsCollisionOfPolygon::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsCollisionOfPolygon::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsCollisionOfPolygon();
	YsCollisionOfPolygon(const YsCollisionOfPolygon &from);
	virtual ~YsCollisionOfPolygon();

	void SetPolygon1(int nv1,const YsVec3 v1[]);
	void SetPolygon2(int nv2,const YsVec3 v2[]);
	YSBOOL CheckOnTheSamePlane(void) const;
	YSBOOL CheckCollision(void) const;  // Just check true/false
	YSINTERSECTION CheckPenetration(void) const;  // Check aprt/touch/penetrate

	YSBOOL OneEdgeLiesOnTheOtherPlane(void) const;

protected:
	int np1;
	const YsVec3 *p1;
	YsVec3 p1min,p1max;
	YsPlane pln1;

	int np2;
	const YsVec3 *p2;
	YsVec3 p2min,p2max;
	YsPlane pln2;
};

// Declaration /////////////////////////////////////////////
class YsSwordPolygonAttr
{
public:
	YsSwordPolygonAttr();
	YSBOOL isOverlappingPolygon;  // For finding overlapping polygon
	int noteOverlapping;          // for polygon ID or any purpose
};

class YsSwordPolygon : public YsBaseClass
{
public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsSwordPolygon::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsSwordPolygon::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsSwordPolygon();
	YsSwordPolygon(const YsSwordPolygon &from);
	virtual ~YsSwordPolygon();

	YsSwordPolygonAttr attrSet;

	YsArray <YsVec3> plg;  // Original Polygon
	YsArray <YsVec2> prj;  // Projected Polygon
	YsArray <int> vtId;    // Corresponding Vertex Id (option)
};

// Declaration /////////////////////////////////////////////
class YsSwordNewVertexLog : public YsBaseClass
{
public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsSwordNewVertexLog::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsSwordNewVertexLog::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsSwordNewVertexLog();
	YsSwordNewVertexLog(const YsSwordNewVertexLog &from);
	virtual ~YsSwordNewVertexLog();

	int freeAttribute;
	int betweenVt1,betweenVt2;
	int idCreated;
	YsVec3 pos;
};

// Declaration /////////////////////////////////////////////
class YsSword : public YsWizard
{
private:
	const YsSword &operator=(const YsSword &from);  // Don't use copy operator
	YsSword(const YsSword &from); // Don't use copy constructor
public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsSword::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsSword::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsSword();
	virtual ~YsSword();

	int GetNumPolygon(void) const;
	int GetNumVertexOfPolygon(int id) const;
	const YsArray <YsVec3> *GetPolygon(int id) const;
	const YsArray <int> *GetVertexIdList(int id) const;

	int GetNumNewVertex(void) const;
	const YsSwordNewVertexLog *GetNewVertexLog(int i) const;

	// For 3D Operation (Slash by plane)
	YSRESULT SetInitialPolygon(int np,const YsVec3 p[],const int id[]=NULL);
	YSRESULT Slash(const YsPlane &pln);
	YSRESULT Slash(int nv,const YsVec3 v[]);
	YSRESULT SlashByOverlappingPolygon3(int nv,const YsVec3 v[],int note=0);
	YSBOOL IsFromOverlappingPolygon3(int i) const;
	YSBOOL IsFromOverlappingPolygon3(int i,int &note) const;
	YSRESULT GetVertexListOfPolygon
	    (YsVec3 vec[],int maxcount,int id) const;
	YSRESULT GetVertexListOfPolygon
	    (const YsVec3 *vec[],int maxcount,int id) const;

	// For 2D Operation (Slash by line)
	YSRESULT SetInitialPolygon(int np,const YsVec2 p[],const int id[]=NULL);
	YSRESULT Slash(const YsVec2 &org,const YsVec2 &vec);
	YSRESULT GetVertexListOfPolygon
	    (YsVec2 vec[],int maxcount,int id) const;

	// For both
	YSRESULT Triangulate(YSCONVEXNIZESTRATEGY strategy=YSCONVEXNIZEDEFAULT);
	YSRESULT Convexnize(YSCONVEXNIZESTRATEGY strategy=YSCONVEXNIZEDEFAULT);

protected:
	YsList <YsSwordPolygon> *SeparateByTwo
	   (YsList <YsSwordPolygon> *org,int id1,int id2);

	YsList <YsSwordPolygon> *SeparateByMulti
	    (YsList <YsSwordPolygon> *org,YsArray <int> &cutPoint);


	YSRESULT SlashOnePolygon
	    (YsList <YsSwordPolygon> *target,
	     const YsPlane &pln,int nv=0,const YsVec3 v[]=NULL);
	YSRESULT FindAddSlashPoint
	    (YsArray <int> &idx,YsList <YsSwordPolygon> *taget,const YsPlane &pln,
	     int nCuttingPolygonVtx=0,const YsVec3 *cuttingPolygon=NULL);

	YSRESULT ConvexnizeGo
	    (YSBOOL tri,
	     YsList <YsSwordPolygon> *target,
	     YSCONVEXNIZESTRATEGY strategy);
	YsList <YsSwordPolygon> *YsSword::ConvexnizeOnePolygon
	    (YSBOOL tri,
	     YsList <YsSwordPolygon> *target,
	     YSCONVEXNIZESTRATEGY strategy);

	YSRESULT FindConvexnizeVertexPair
	    (int &id1,int &id2,
	     YsList <YsSwordPolygon> *target,YSCONVEXNIZESTRATEGY strategy);

	int newCreateId;
	YsMatrix4x4 plgPrjMatrix;
	YsList <YsSwordPolygon> *plg;
	YsList <YsSwordNewVertexLog> *cre;

	YSRESULT Separate
		   (YsArray <YsVec3> &plg1,
		    YsArray <YsVec3> &plg2,
		    YsArray <YsVec3> &org,
		    int id1,int id2);
	YSRESULT Separate
		   (YsArray <YsVec2> &plg1,
		    YsArray <YsVec2> &plg2,
		    YsArray <YsVec2> &org,
		    int id1,int id2);
	YSRESULT Separate
		   (YsArray <int> &vtId1,
		    YsArray <int> &vtId2,
		    YsArray <int> &org,
		    int id1,int id2);
};

// Declaration /////////////////////////////////////////////
class YsBoundingBoxMaker2 : public YsWizard
{
public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsBoundingBoxMaker2::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsBoundingBoxMaker2::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsBoundingBoxMaker2();
	YsBoundingBoxMaker2(const YsBoundingBoxMaker2 &from);
	virtual ~YsBoundingBoxMaker2();

	void Begin(const YsVec2 &vec);
	void Add(const YsVec2 &vec);
	void Make(int n,const YsVec2 v[]);
	void Make(YsList <YsVec2> *lst);
	void Make(const YsArray <YsVec2> &ary);
	void Get(YsVec2 &min,YsVec2 &max) const;
protected:
	YsVec2 min,max;
};

// Declaration /////////////////////////////////////////////
class YsBoundingBoxMaker3 : public YsWizard
{
public:
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsBoundingBoxMaker3::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsBoundingBoxMaker3::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsBoundingBoxMaker3();
	YsBoundingBoxMaker3(const YsBoundingBoxMaker3 &from);
	virtual ~YsBoundingBoxMaker3();

	void Begin(const YsVec3 &vec);
	void Add(const YsVec3 &vec);
	void Make(int n,const YsVec3 v[]);
	void Make(const YsList <YsVec3> *lst);
	void Make(const YsArray <YsVec3> &ary);
	void Get(YsVec3 &min,YsVec3 &max) const;
protected:
	YsVec3 min,max;
};

/* } */
#endif
