#include <stdio.h>
#include <string.h>
#include <stdlib.h>


#include "ysclass.h"

////////////////////////////////////////////////////////////

YSBOOL YsCheckConvexByAngle2(int np,const YsVec2 p[])
{
	int i;
	const YsVec2 *p1,*p2;
	YsVec2 v1,v2;
	double o1,o2;

	while(np>3 && p[0]==p[1])
	{
		p++;
		np--;
	}
	while(np>3 && p[np-1]==p[np-2])
	{
		np--;
	}
	if(np>3 && p[np-1]==p[0])
	{
		np--;
	}

	if(np<=3)
	{
		return YSTRUE;
	}

 	p1=&p[np-2];
	p2=&p[np-1];
	v2=*p2-*p1;
	for(i=0; i<np; i++)
	{
		p1=p2;
		p2=&p[i];
		if(*p1!=*p2)
		{
			v1=v2;
			v2=*p2-*p1;

			o1=v1^v2;
			if(YsAbs(o1)>YsTolerance)
			{
				break;
			}
		}
	}

	for(i=i+1; i<np; i++)
	{
		p1=p2;
		p2=&p[i];
		if(*p1!=*p2)
		{
			v1=v2;
			v2=*p2-*p1;

			o2=v1^v2;
			if(YsAbs(o2)>YsTolerance && o1*o2<0)
			{
				return YSFALSE;
			}
		}
	}
	return YSTRUE;
}

YSBOOL YsCheckConvexByAngle3(int np,const YsVec3 p[])
{
	int i;
	const YsVec3 *p1,*p2;
	YsVec3 v1,v2,o1,o2;

	while(np>3 && p[0]==p[1])
	{
		p++;
		np--;
	}
	while(np>3 && p[np-1]==p[np-2])
	{
		np--;
	}
	if(np>3 && p[np-1]==p[0])
	{
		np--;
	}

	if(np<=3)
	{
		return YSTRUE;
	}

 	p1=&p[np-2];
	p2=&p[np-1];
	v2=*p2-*p1;
	for(i=0; i<np; i++)
	{
		p1=p2;
		p2=&p[i];
		if(*p1!=*p2)
		{
			v1=v2;
			v2=*p2-*p1;

			o1=v1^v2;
			if(o1.GetSquareLength()>YsTolerance)
			{
				break;
			}
		}
	}

	for(i=i+1; i<np; i++)
	{
		p1=p2;
		p2=&p[i];
		if(*p1!=*p2)
		{
			v1=v2;
			v2=*p2-*p1;

			o2=v1^v2;
			if(o2.GetSquareLength()>YsTolerance && o1*o2<0)
			{
				return YSFALSE;
			}
		}
	}

	return YSTRUE;
}

YSBOOL YsCheckConvex2(int np,const YsVec2 p[],YSBOOL strictCheck)
{
	if(strictCheck!=YSTRUE)
	{
		return YsCheckConvexByAngle2(np,p);
	}
	else
	{
		int i,j;

		if(np<=3)
		{
			return YSTRUE;
		}

		if(YsCheckConvexByAngle2(np,p)==YSTRUE)
		{
			for(i=0; i<np; i++)
			{
				double z;
				int idx0,idx1,idx2;

				idx0=(i+np-1)%np;
				idx1=i;
				idx2=(i+1)%np;

				z=(p[idx1]-p[idx0])^(p[idx2]-p[idx1]);

				if(z*z<YsTolerance)
				{
					continue;
				}

				for(j=0; j<np; j++)
				{
					if( j==i-1 ||
					    j==i+1 ||
					   (i==0 && j==np-1) ||
					   (i==np-1 && j==0))
					{
						continue;      // Skip neighbor
					}

					idx0=(j+np-1)%np;
					idx1=j;
					idx2=(j+1)%np;

					z=(p[idx1]-p[idx0])^(p[idx2]-p[idx1]);

					if(z*z<YsTolerance)
					{
						continue;     // Skip Straight lines case
					}

					if(YsCheckSeparatability2(np,p,i,j)!=YSTRUE)
					{
						goto CONCAVE;
					}
				}
			}
			return YSTRUE;
		}
	CONCAVE:
		return YSFALSE;
	}
}

YSBOOL YsCheckConvex3(int np,const YsVec3 p[],YSBOOL strictCheck)
{
	if(strictCheck!=YSTRUE)
	{
		return YsCheckConvexByAngle3(np,p);
	}
	else
	{
		int i,j;

		if(np<=3)
		{
			return YSTRUE;
		}

		if(YsCheckConvexByAngle3(np,p)==YSTRUE)
		{
			for(i=0; i<np; i++)
			{
				YsVec3 z;
				int idx0,idx1,idx2;

				idx0=(i+np-1)%np;
				idx1=i;
				idx2=(i+1)%np;

				z=(p[idx1]-p[idx0])^(p[idx2]-p[idx1]);

				if(z.GetSquareLength()<YsTolerance)
				{
					continue;
				}

				for(j=0; j<np; j++)
				{
					if( j==i-1 ||
					    j==i+1 ||
					   (i==0 && j==np-1) ||
					   (i==np-1 && j==0))
					{
						continue;      // Skip neighbor
					}

					idx0=(j+np-1)%np;
					idx1=j;
					idx2=(j+1)%np;

					z=(p[idx1]-p[idx0])^(p[idx2]-p[idx1]);

					if(z.GetSquareLength()<YsTolerance)
					{
						continue;     // Skip Straight lines case
					}

					if(YsCheckSeparatability3(np,p,i,j)!=YSTRUE)
					{
						goto CONCAVE;
					}
				}
			}
			return YSTRUE;
		}
	CONCAVE:
		return YSFALSE;
	}
}

////////////////////////////////////////////////////////////

double YsGetTriangleArea3(const YsVec3 &p1,const YsVec3 &p2,const YsVec3 &p3)
{
	YsVec3 a,b;

	a=p2-p1;
	b=p3-p1;

//	double c,s,la,lb;
//	la=a.GetLength();
//	lb=b.GetLength();
//	c=a*b/(la*lb);
//	s=sqrt(1.0-c*c);
//	return la*lb*s/2.0;

	return (a^b).GetLength()/2.0;
}

////////////////////////////////////////////////////////////

double YsGetTriangleArea2(const YsVec2 &p1,const YsVec2 &p2,const YsVec2 &p3)
{
	YsVec2 a,b;
	a=p2-p1;
	b=p3-p1;
	return (a^b)/2.0;
}

////////////////////////////////////////////////////////////

YSRESULT YsGetLargestTriangleFromPolygon3
    (const YsVec3 *v[3],int np,const YsVec3 p[])
{
	int i,j,k;
	double candidate,current;
	if(np==3)
	{
		v[0]=&p[0];
		v[1]=&p[1];
		v[2]=&p[2];
		return YSOK;
	}
	else if(np>=3)
	{
		candidate=0.0;
		v[0]=&p[0];
		v[1]=&p[1];
		v[2]=&p[2];
		for(i=0; i<np; i++)
		{
			for(j=i+1; j<np; j++)
			{
				for(k=j+1; k<np; k++)
				{
					current=((p[j]-p[i])^(p[k]-p[i])).GetSquareLength();
					if(current>candidate)
					{
						candidate=current;
						v[0]=&p[i];
						v[1]=&p[j];
						v[2]=&p[k];
					}
				}
			}
		}
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsGetLargestTriangleFromPolygon3(YsVec3 v[3],int np,const YsVec3 p[])
{
	const YsVec3 *tri[3];
	if(YsGetLargestTriangleFromPolygon3(tri,np,p)==YSOK)
	{
		v[0]=*tri[0];
		v[1]=*tri[1];
		v[2]=*tri[2];
		return YSOK;
	}
	return YSERR;
}

////////////////////////////////////////////////////////////

YSRESULT YsGetLargestTriangleFromPolygon2
    (const YsVec2 *v[3],int np,const YsVec2 p[])
{
	int i,j,k;
	double candidate,current;

	if(np==3)
	{
		v[0]=&p[0];
		v[1]=&p[1];
		v[2]=&p[2];
		return YSOK;
	}
	else if(np>3)
	{
		candidate=0.0;
		v[0]=&p[0];
		v[1]=&p[1];
		v[2]=&p[2];
		for(i=0; i<np; i++)
		{
			for(j=i+1; j<np; j++)
			{
				for(k=j+1; k<np; k++)
				{
					current=((p[j]-p[i])^(p[k]-p[i]));
					if(current>candidate)
					{
						candidate=current;
						v[0]=&p[i];
						v[1]=&p[j];
						v[2]=&p[k];
					}
				}
			}
		}
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsGetLargestTriangleFromPolygon2(YsVec2 v[3],int np,const YsVec2 p[])
{
	const YsVec2 *tri[3];
	if(YsGetLargestTriangleFromPolygon2(tri,np,p)==YSOK)
	{
		v[0]=*tri[0];
		v[1]=*tri[1];
		v[2]=*tri[2];
		return YSOK;
	}
	return YSERR;
}

////////////////////////////////////////////////////////////

YSRESULT YsGetAverageNormalVector(YsVec3 &nom,int np,const YsVec3 p[])
{
	if(np==3)
	{
		nom=(p[1]-p[0])^(p[2]-p[1]);
		if(nom.GetSquareLength()>=YsTolerance)
		{
			nom.Normalize();
			return YSOK;
		}
	}
	else
	{
		const YsVec3 *tri[3];
		if(YsGetLargestTriangleFromPolygon3(tri,np,p)==YSOK)
		{
			nom=(*tri[1]-*tri[0])^(*tri[2]-*tri[1]);
			if(nom.GetSquareLength()>=YsTolerance)
			{
				nom.Normalize();
				return YSOK;
			}
		}
	}
	return YSERR;
}

////////////////////////////////////////////////////////////


/************************************************************************

   1)y-y1=dy/dx*(x-x1)
   2)dx(y-y1)=dy(x-x1)
   3)dx*y-dx*y1=dy*x-dy*x1
   4)dy*x-dx*y-dy*x1+dx*y1=0

   ax+by+c=0
     a= dy
     b=-dx
     c=-dy*x1+dx*y1

   a1 x  +b1 y  +c1 = 0
   a2 x  +b2 y  +c2 = 0

   a2a1 x  +a2b1 y  +a2c1 = 0
   a1a2 x  +a1b2 y  +a1c2 = 0
  ----------------------------
         (a2b1-a1b2)y +a2c1-a1c2 = 0


   a1b2 x  +b1b2 y  +c1b2 = 0
   a2b1 x  +b2b1 y  +c2b1 = 0
  ----------------------------
   (a1b2-a2b1)x +c1b2-c2b1 = 0

************************************************************************/

static void YsGetLineCoeffs
    (double *a,double *b,double *c,const YsVec2 &o,const YsVec2 &v)
{
	*a= v.y();
	*b=-v.x();
	*c=-v.y()*o.x() +v.x()*o.y();
}

YSRESULT YsGetLineIntersection2
  (YsVec2 &crs,
   const YsVec2 &p1,const YsVec2 &p2,const YsVec2 &q1,const YsVec2 &q2)
{
	double a1,b1,c1;
	double a2,b2,c2;
	double p,q;
	double x,y;
	YsVec2 vp,vq;

	vp=p2-p1;
	vq=q2-q1;

	YsGetLineCoeffs(&a1,&b1,&c1,p1,vp);
	YsGetLineCoeffs(&a2,&b2,&c2,q1,vq);

	p=a2*b1-a1*b2;
	q=a2*c1-a1*c2;
	if(YsEqual(p,0.0)==YSTRUE)
	{
		return YSERR;
	}
	y=-q/p;

	p=a1*b2-a2*b1;
	q=c1*b2-c2*b1;
	if(YsEqual(p,0.0)==YSTRUE)
	{
		return YSERR;
	}
	x=-q/p;

	crs.Set(x,y);
	return YSOK;
}

YSBOOL YsCheckLineOverlap2
    (const YsVec2 &o1,const YsVec2 &v1,const YsVec2 &o2,const YsVec2 &v2)
{
	if(YsAbs(v1^v2)<YsTolerance &&
	   YsAbs((o2-o1)^v1)<YsTolerance &&
	   v1.GetSquareLength()>YsTolerance &&
	   v2.GetSquareLength()>YsTolerance)
	{
		return YSTRUE;
	}
	return YSFALSE;


/* Comparing Coefficient method
	double a1,b1,c1;
	double a2,b2,c2;
	double p,q;

	YsGetLineCoeffs(&a1,&b1,&c1,o1,v1);
	YsGetLineCoeffs(&a2,&b2,&c2,o2,v2);

	p=a2*b1-a1*b2;
	q=a2*c1-a1*c2;
	if(YsEqual(p,0.0)==YSTRUE && YsEqual(q,0.0)==YSTRUE)
	{
		return YSTRUE;
	}

	p=a1*b2-a2*b1;
	q=c1*b2-c2*b1;
	if(YsEqual(p,0.0)==YSTRUE && YsEqual(q,0.0)==YSTRUE)
	{
		return YSTRUE;
	}

	return YSFALSE;
*/
}

YSBOOL YsCheckLineOverlap3
    (const YsVec3 &o1,const YsVec3 &v1,const YsVec3 &o2,const YsVec3 &v2)
{
	if((v1^v2).GetSquareLength()<YsTolerance &&
	   ((o2-o1)^v1).GetSquareLength()<YsTolerance &&
	   v1.GetSquareLength()>YsTolerance &&
	   v2.GetSquareLength()>YsTolerance)
	{
		return YSTRUE;
	}
	return YSFALSE;
}

YSSIDE YsCheckInsidePolygon2(const YsVec2 &pnt,int np,const YsVec2 p[])
{
	int i,j,CrossCounter,vt1,vt2;
	double ang;
	YsVec2 vec,crs,cVec;
	YsVec2 o1,v1,o2,v2;

	ang=0.0;
	o1=pnt;
	for(i=0; i<256; i++)
	{
		v1.Set(cos(ang),sin(ang));

		CrossCounter=0;
		for(j=0; j<np; j++)
		{
			vt1= j;
			vt2=(j+1)%np;

			o2=p[vt1];
			v2=p[vt2]-p[vt1];

			if(pnt==p[vt1] || pnt==p[vt2])
			{
				return YSBOUNDARY;
			}

			if(p[vt1]==p[vt2])
			{
				continue;
			}

			if(YsGetLineIntersection2(crs,o1,o1+v1,o2,o2+v2)==YSOK)
			{
				if(crs==pnt && YsCheckInBetween2(crs,p[vt2],p[vt1])==YSTRUE)
				{
					return YSBOUNDARY;
				}
				if(crs==p[vt1] || crs==p[vt2])
				{
					goto NEXTANGLE;
				}

				cVec=crs-pnt;
				if(cVec*v1>0.0 && YsCheckInBetween2(crs,p[vt1],p[vt2])==YSTRUE)
				{
					CrossCounter++;
				}
			}
			else if(YsCheckLineOverlap2(o1,v1,o2,v2)==YSTRUE)
			{
				goto NEXTANGLE;
			}
		}
		if(CrossCounter%2==0)
		{
			return YSOUTSIDE;
		}
		else
		{
			return YSINSIDE;
		}

	NEXTANGLE:
		ang+=(92.0*YsPi/180.0);
	}
	YsPrintf::def->Printf("YsCheckInsidePolyg2\nWarning\n");
	return YSUNKNOWNSIDE;
}





YSRESULT YsGetCenterOfPolygon3(YsVec3 &cen,int np,const YsVec3 plg[])
{
	const YsVec3 *tri[3];
	if(YsGetLargestTriangleFromPolygon3(tri,np,plg)==YSOK)
	{
		cen=(*tri[0]+*tri[1]+*tri[2])/3.0;
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsGetCenterOfPolygon2(YsVec2 &cen,int np,const YsVec2 plg[])
{
	const YsVec2 *tri[3];
	if(YsGetLargestTriangleFromPolygon2(tri,np,plg)==YSOK)
	{
		cen=(*tri[0]+*tri[1]+*tri[2])/3.0;
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsGetPolygonProjectionMatrix(YsMatrix4x4 &mat,int np,const YsVec3 p[])
{
	YsVec3 tri[3],cen,nom;
	if(YsGetLargestTriangleFromPolygon3(tri,np,p)==YSOK)
	{
		mat.Initialize();

		nom=((tri[1]-tri[0])^(tri[2]-tri[1]));
		nom.Normalize();
		cen=((tri[0]+tri[1]+tri[2])/3.0);

		double hdg,pch;
		hdg=atan2(nom.x(),-nom.z());
		pch=asin(nom.y());
		mat.RotateZY( pch);
		mat.RotateXZ(-hdg);

		mat.Translate(-cen);
		return YSOK;
	}
	return YSERR;
}

YSINTERSECTION YsGetLinePenetration2
    (const YsVec2 &p1,const YsVec2 &p2,const YsVec2 &q1,const YsVec2 &q2)
{
	if(p1==q1 || p1==q2 || p2==q1 || p2==q2)
	{
		return YSSHAREPOINT;
	}
	else
	{
		// First, let's check bounding box
		double x1,y1,x2,y2;
		x1=YsSmaller(p1.x(),p2.x());
		y1=YsSmaller(p1.y(),p2.y());
		x2=YsGreater(p1.x(),p2.x());
		y2=YsGreater(p1.y(),p2.y());

		if((q1.x()<x1 && q2.x()<x1)||(q1.x()>x2 && q2.x()>x2)||
		   (q1.y()<y1 && q2.y()<y1)||(q1.y()>y2 && q2.y()>y2))
		{
			return YSAPART;
		}

		YsVec2 crs;
		if(YsGetLineIntersection2(crs,p1,p2,q1,q2)==YSOK)
		{
			if((YsCheckInBetween2(crs,p1,p2)==YSTRUE&&(q1==crs||q2==crs))||
			   (YsCheckInBetween2(crs,q1,q2)==YSTRUE&&(p1==crs||p2==crs)))
			{
				return YSTOUCH;
			}
			else if(YsCheckInBetween2(crs,p1,p2)==YSTRUE &&
			        YsCheckInBetween2(crs,q1,q2)==YSTRUE)
			{
				return YSINTERSECT;
			}
			else
			{
				return YSAPART;
			}
		}
		else if(YsCheckLineOverlap2(p1,p2-p1,q1,q2-q1)==YSTRUE)
		{
			return YSOVERLAP;
		}
		else
		{
			return YSAPART;
		}
	}
}

YSINTERSECTION YsGetLinePenetration3
    (const YsVec3 &p1,const YsVec3 &p2,const YsVec3 &q1,const YsVec3 &q2)
{
	if(p1==q1 || p1==q2 || p2==q1 || p2==q2)
	{
		return YSSHAREPOINT;
	}
	else
	{
		// First, let's check bounding box
		double x1,y1,z1,x2,y2,z2;
		x1=YsSmaller(p1.x(),p2.x());
		y1=YsSmaller(p1.y(),p2.y());
		z1=YsSmaller(p1.z(),p2.z());
		x2=YsGreater(p1.x(),p2.x());
		y2=YsGreater(p1.y(),p2.y());
		z2=YsGreater(p1.z(),p2.z());

		if((q1.x()<x1 && q2.x()<x1)||(q1.x()>x2 && q2.x()>x2)||
		   (q1.y()<y1 && q2.y()<y1)||(q1.y()>y2 && q2.y()>y2)||
		   (q1.z()<z1 && q2.z()<z1)||(q1.z()>z2 && q2.z()>z2))
		{
			return YSAPART;
		}

		YsVec3 crs1,crs2;
		if(YsGetNearestPointOfTwoLine(crs1,crs2,p1,p2,q1,q2)==YSOK)
		{
			if((YsCheckInBetween3(crs1,p1,p2)==YSTRUE&&(q1==crs1||q2==crs1))||
			   (YsCheckInBetween3(crs2,q1,q2)==YSTRUE&&(p1==crs2||p2==crs2)))
			{
				return YSTOUCH;
			}
			else if(YsCheckInBetween3(crs1,p1,p2)==YSTRUE &&
			        YsCheckInBetween3(crs2,q1,q2)==YSTRUE)
			{
				return YSINTERSECT;
			}
			else
			{
				return YSAPART;
			}
		}
		else if(YsCheckLineOverlap3(p1,p2-p1,q1,q2-q1)==YSTRUE)
		{
			return YSOVERLAP;
		}
		else
		{
			return YSAPART;
		}
	}
}

YSBOOL YsCheckSeparatability2(int np,const YsVec2 p[],int idx1,int idx2)
{
	if(np<=3 || idx1==idx2 || idx1==(idx2+1)%np || idx2==(idx1+1)%np)
	{
		return YSFALSE;
	}
	else if(0<=idx1 && idx1<np && 0<=idx2 && idx2<np)
	{
		YsVec2 r1,r2,rc,s1,s2;
		r1=p[idx1];
		r2=p[idx2];

		rc=(r1+r2)/2.0;
		if(YsCheckInsidePolygon2(rc,np,p)==YSINSIDE)
		{
			int i;
			for(i=0; i<np; i++)
			{
				YSINTERSECTION is;
				is=YsGetLinePenetration2(p[i],p[(i+1)%np],r1,r2);
				if(is!=YSAPART && is!=YSSHAREPOINT)
				{
					return YSFALSE;
				}
			}
			return YSTRUE;
		}
	}
	return YSFALSE;
}


YSBOOL YsCheckSeparatability3(int np,const YsVec3 p[],int idx1,int idx2)
{
	if(np<=3 || idx1==idx2 || idx1==(idx2+1)%np || idx2==(idx1+1)%np)
	{
		return YSFALSE;
	}
	else if(0<=idx1 && idx1<np && 0<=idx2 && idx2<np)
	{
		YsVec3 r1,r2,rc,s1,s2;
		r1=p[idx1];
		r2=p[idx2];

		rc=(r1+r2)/2.0;
		if(YsCheckInsidePolygon3(rc,np,p)==YSINSIDE)
		{
			int i;
			for(i=0; i<np; i++)
			{
				YSINTERSECTION is;
				is=YsGetLinePenetration3(p[i],p[(i+1)%np],r1,r2);
				if(is!=YSAPART && is!=YSSHAREPOINT)
				{
					return YSFALSE;
				}
			}
			return YSTRUE;
		}
	}
	return YSFALSE;

/* Project to XY plane method
	int i;
	YsVec3 prj;
	YsVec2 *p2d;
	YsMatrix4x4 mat;
	YSBOOL separatability;
	YsOneMemoryBlock <YsVec2,128> memBlock;


	if(idx1==idx2 || idx1==(idx2+1)%np || idx2==(idx1+1)%np)
	{
		return YSFALSE;
	}

	if(YsGetPolygonProjectionMatrix(mat,np,p)==YSOK)
	{
		p2d=memBlock.GetMemoryBlock(np);

		if(p2d!=NULL)
		{
			for(i=0; i<np; i++)
			{
				prj=mat*p[i];
				p2d[i].Set(prj.x(),prj.y());
			}

			separatability=YsCheckSeparatability2(np,p2d,idx1,idx2);

			return separatability;
		}
	}
	return YSFALSE;
*/
}


YSBOOL YsCheckBoundingBoxCollision3
(const YsVec3 &min1,const YsVec3 &max1,const YsVec3 &min2,const YsVec3 &max2)
{
	if(max1.x()<min2.x()-YsTolerance || max2.x()<min1.x()-YsTolerance ||
	   max1.y()<min2.y()-YsTolerance || max2.y()<min1.y()-YsTolerance ||
	   max1.z()<min2.z()-YsTolerance || max2.z()<min1.z()-YsTolerance)
	{
		return YSFALSE;
	}
	return YSTRUE;
}

YSBOOL YsCheckBoundingBoxCollision2
(const YsVec2 &min1,const YsVec2 &max1,const YsVec2 &min2,const YsVec2 &max2)
{
	if(max1.x()<min2.x()-YsTolerance || max2.x()<min1.x()-YsTolerance ||
	   max1.y()<min2.y()-YsTolerance || max2.y()<min1.y()-YsTolerance)
	{
		return YSFALSE;
	}
	return YSTRUE;
}


YSBOOL YsCheckInsideBoundingBox3
    (const YsVec3 &ref,const YsVec3 &min,const YsVec3 &max)
{
	if(min.x()-YsTolerance<=ref.x() && ref.x()<=max.x()+YsTolerance &&
	   min.y()-YsTolerance<=ref.y() && ref.y()<=max.y()+YsTolerance &&
	   min.z()-YsTolerance<=ref.z() && ref.z()<=max.z()+YsTolerance)
	{
		return YSTRUE;
	}
	return YSFALSE;
}

YSBOOL YsCheckInsideBoundingBox2
    (const YsVec2 &ref,const YsVec2 &min,const YsVec2 &max)
{
	if(min.x()-YsTolerance<=ref.x() && ref.x()<=max.x()+YsTolerance &&
	   min.y()-YsTolerance<=ref.y() && ref.y()<=max.y()+YsTolerance)
	{
		return YSTRUE;
	}
	return YSFALSE;
}


YSFLIPDIRECTION YsCheckFlipDirection2(int np,const YsVec2 p[])
{
	int i,j;
	for(i=0; i<np; i++)
	{
		const YsVec2 *vt1,*vt2;
		vt1=&p[i];
		vt2=&p[(i+1)%np];
		if(*vt1==*vt2)
		{
			continue;
		}

		for(j=1; j<4; j++)
		{
			int k,nCrs;
			YsVec2 c,t,n;
			t=(*vt2-*vt1);
			c=*vt1+t*double(j)/4.0;
			n.Set(t.y(),-t.x());

			nCrs=0;
			for(k=0; k<np; k++)
			{
				const YsVec2 *r1,*r2;
				YsVec2 is;
				r1=&p[k];
				r2=&p[(k+1)%np];

				if(k==i || *r1==*r2)
				{
					continue;
				}

				if((*r2-c)*n<0.0 && (*r1-c)*n<0.0)
				{
					continue;
				}

				if(YsGetLineIntersection2(is,c,c+n,*r1,*r2)==YSOK)
				{
					if(is==c || is==*r1 || is==*r2)
					{
						nCrs=-1;
						break;
					}
					if(YsCheckInBetween2(is,*r1,*r2)==YSTRUE &&
					   (is-c)*n>=0.0)
					{
						nCrs++;
					}
				}
				else if(YsCheckLineOverlap2(*vt1,*vt2-*vt1,*r1,*r2-*r1)==YSTRUE)
				{
					nCrs=-1;
					break;
				}
			}
			if(nCrs>=0 && nCrs%2!=0)
			{
				return YSFLIPCW;
			}
			else if(nCrs>=0 && nCrs%2==0)
			{
				return YSFLIPCCW;
			}
		}
	}
	return YSFLIPUNKNOWN;
}

YSFLIPDIRECTION YsCheckFlipDirection3(int np,const YsVec3 p[],const YsVec3 &n)
{
	int i;
	double hdg,pch;
	YsVec3 nom;
	YsVec2 *q;
	YsMatrix4x4 mat;
	YSFLIPDIRECTION flip;

	YsOneMemoryBlock <YsVec2,64> memBlock;

	if(np>=3 && n.GetSquareLength()>YsTolerance)
	{
		nom=n;
		nom.Normalize();

		hdg=atan2(nom.x(),-nom.z());
		pch=asin(nom.y());
		mat.Initialize();
		mat.RotateZY( pch);
		mat.RotateXZ(-hdg);

		q=memBlock.GetMemoryBlock(np);

		if(q!=NULL)
		{
			for(i=0; i<np; i++)
			{
				YsVec3 prj;
				prj=mat*p[i];
				q[i].Set(prj.x(),prj.y());
			}
			flip=YsCheckFlipDirection2(np,q);

			if(YsCoordSysModel==YSLEFT_ZPLUS_YPLUS)
			{
				if(flip==YSFLIPCCW)
				{
					flip=YSFLIPCW;
				}
				else if(flip==YSFLIPCW)
				{
					flip=YSFLIPCCW;
				}
			}
			else if(YsCoordSysModel==YSRIGHT_ZMINUS_YPLUS)
			{
				// Right Hand -> Don't have to be inversed.
			}
			else
			{
				YsPrintf::def->Printf(
					"YsCheckFlipDirection3()\n"
					"Unknown Coordinate System Model\n"
					"Assuming Right hand\n");
			}

			return flip;
		}
		else
		{
			YsPrintf::def->Printf(
			    "YsCheckFlipDirection3()\nLowMemoryWarning!!\n");
		}
	}
	return YSFLIPUNKNOWN;

}


YSRESULT YsGetTwoPlaneCrossLine
    (YsVec3 &org,YsVec3 &vec,const YsPlane &pl1,const YsPlane &pl2)
{
	/******************************************************************
	Line Direction Vector 'lin->v' is Outer Product of 'pl1->n','pl2->n'.

	'p1l' is the line that crosses pl1->o and whose direction vector is
	the Outer Product of 'pl1->n' and 'lin->v'. The point which 'p1l' and
	'pl2' cross will be 'lin->o'.
	******************************************************************/

	YsVec3 org1,nom1,org2,nom2;
	pl1.GetOrigin(org1);
	pl1.GetNormal(nom1);
	pl2.GetOrigin(org2);
	pl2.GetNormal(nom2);

	vec=nom1^nom2;
	if(vec.GetSquareLength()>YsTolerance)
	{
		vec.Normalize();

		YsVec3 guide;
		guide=nom1^vec;

		if(guide.GetSquareLength()>YsTolerance)
		{
			guide.Normalize();
			if(pl2.GetIntersection(org,org1,guide)==YSOK)
			{
				return YSOK;
			}
		}
	}
	return YSERR;
}


YSRESULT YsGetNearestPointOfTwoLine
    (YsVec3 &np,YsVec3 &nq,
     const YsVec3 &p1,const YsVec3 &p2,const YsVec3 &q1,const YsVec3 &q2)
{
	double a,b,c,d,e,f,s,t;
	YsVec3 pv,qv;

	pv=p2-p1;
	qv=q2-q1;

	a= pv*pv;
	b=-qv*pv;
	c= p1*pv-q1*pv;
	d= pv*qv;
	e=-qv*qv;
	f= p1*qv-q1*qv;

	if(YsSolveTwoLinearEquation(&t,&s,a,b,c,d,e,f)==YSOK)
	{
		np=p1+t*pv;
		nq=q1+s*qv;
		return YSOK;
	}
	else
	{
		np=p1;
		YsGetNearestPointOnLine3(nq,q1,q2,np);
	}
	return YSERR;
}

YSRESULT YsGetNearestPointOnLine3
    (YsVec3 &np,const YsVec3 &p1,const YsVec3 &p2,const YsVec3 &ref)
{
	double l;
	YsVec3 pv,rv;

	pv=p2-p1;
	if(pv.GetSquareLength()>YsTolerance)
	{
		pv.Normalize();
		rv=ref-p1;
		l=rv*pv;
		np=p1+pv*l;
		return YSOK;
	}
	np=p1;
	return YSERR;
}

YSRESULT YsGetNearestPointOnLine2
    (YsVec2 &np,const YsVec2 &p1,const YsVec2 &p2,const YsVec2 &ref)
{
	double l;
	YsVec2 pv,rv;

	pv=p2-p1;
	if(pv.GetSquareLength()>YsTolerance)
	{
		pv.Normalize();
		rv=ref-p1;
		l=rv*pv;
		np=p1+pv*l;
		return YSOK;
	}
	np=p1;
	return YSERR;
}


static YSINTERSECTION YsGetPolygonEdgeRelation
(int np,const YsVec3 p[],const YsPlane &pln,const YsVec3 &e1,const YsVec3 &e2)
{
	if(e1!=e2 &&
	   pln.CheckOnPlane(e1)==YSTRUE &&
	   pln.CheckOnPlane(e2)==YSTRUE)
	{
		YSSIDE side1,side2;
		side1=YsCheckInsidePolygon3(e1,np,p);
		side2=YsCheckInsidePolygon3(e2,np,p);
		if(side1==YSINSIDE || side2==YSINSIDE)
		{
			return YSINTERSECT;
		}
		else if(side1==YSBOUNDARY || side2==YSBOUNDARY)
		{
			return YSTOUCH;
		}
		else
		{
			if(side1==YSUNKNOWNSIDE ||
			   side2==YSUNKNOWNSIDE)
			{
				YsPrintf::def->Printf("YsGetPolygonEdgeRelation\n"
				                      "  Seems bad polygon(1)\n");
			}

			// Even if it misses the case like
			//     e1    |(polygon boundary)|     e2
			// Another polygon will hit this case.
			return YSAPART;
		}
	}

	// OK, the edge is not parallel to the polygon.

	if(pln.CheckOnPlane(e1)==YSTRUE)
	{
		YSSIDE side1;
		side1=YsCheckInsidePolygon3(e1,np,p);
		if(side1==YSINSIDE || side1==YSBOUNDARY)
		{
			return YSTOUCH;
		}
		if(side1==YSUNKNOWNSIDE)
		{
			YsPrintf::def->Printf("YsGetPolygonEdgeRelation\n"
			                "  Seems bad polygon(2)\n");
		}
	}
	if(pln.CheckOnPlane(e2)==YSTRUE)
	{
		YSSIDE side2;
		side2=YsCheckInsidePolygon3(e2,np,p);
		if(side2==YSINSIDE || side2==YSBOUNDARY)
		{
			return YSTOUCH;
		}
		if(side2==YSUNKNOWNSIDE)
		{
			YsPrintf::def->Printf("YsGetPolygonEdgeRelation\n"
			                "  Seems bad polygon(3)\n");
		}
	}

	// OK, at this point, both end points are not touching the plane.

	if(e1==e2)
	{
		return YSAPART;
	}

	// Not parallel, end points are not touching, the edge has length.

	YsVec3 crs;
	if(pln.GetPenetration(crs,e1,e2)==YSOK)
	{
		YSSIDE side;
		side=YsCheckInsidePolygon3(crs,np,p);
		if(side==YSINSIDE)
		{
			return YSINTERSECT;
		}
		else if(side==YSBOUNDARY)
		{
			return YSTOUCH;
		}
		else if(side==YSUNKNOWNSIDE)
		{
			YsPrintf::def->Printf("YsGetPolygonEdgeRelation\n"
			                "  Seems bad polygon(4)\n");
		}
	}

	return YSAPART;
}


/* Use all edges and separatable combinations to check penetration */
YSINTERSECTION YsGetPolygonPenetration
    (int np,const YsVec3 p[],int nq,const YsVec3 q[])
{
	YsPlane pln;
	YSINTERSECTION isect;
	int touchCount1,touchCount2;
	int i;


	pln.MakeBestFitPlane(np,p);
	touchCount1=0;
	for(i=0; i<nq; i++)
	{
		isect=YsGetPolygonEdgeRelation(np,p,pln,q[i],q[(i+1)%nq]);
		if(isect==YSINTERSECT)
		{
			return YSINTERSECT;
		}
		else if(isect==YSTOUCH)
		{
			touchCount1++;
		}
	}
	if(touchCount1>=2)
	{
		return YSINTERSECT;
	}



	pln.MakeBestFitPlane(nq,q);
	touchCount2=0;
	for(i=0; i<np; i++)
	{
		isect=YsGetPolygonEdgeRelation(nq,q,pln,p[i],p[(i+1)%np]);
		if(isect==YSINTERSECT)
		{
			return YSINTERSECT;
		}
		else if(isect==YSTOUCH)
		{
			touchCount2++;
		}
	}
	if(touchCount2>=2)
	{
		return YSINTERSECT;
	}


	if(touchCount1>0 || touchCount2>0)
	{
		return YSTOUCH;
	}


	return YSAPART;
}

////////////////////////////////////////////////////////////

YSRESULT YsGetArbitraryInsidePointOfPolygon2
     (YsVec2 &aip,int np,const YsVec2 p[],YSBOOL strictConvexityCheck)
{
	if(YsCheckConvex2(np,p,strictConvexityCheck)==YSTRUE)
	{
		int i;
		YsVec2 cg;
		cg.Set(0.0,0.0);
		for(i=0; i<np; i++)
		{
			cg=cg+p[i];
		}
		cg=cg/double(np);
		aip=cg;
		return YSOK;
	}
	else
	{
		int i,j;
		for(i=0; i<np; i++)
		{
			for(j=i+1; j<np; j++)
			{
				if(YsCheckSeparatability2(np,p,i,j)==YSTRUE)
				{
					aip=(p[i]+p[j])/2.0;
					goto DONE;
				}
			}
		}
		YsGetCenterOfPolygon2(aip,np,p);  // I can do nothing at this point(T_T)
		return YSERR;
DONE:
		return YSOK;
	}
}

YSRESULT YsGetArbitraryInsidePointOfPolygon3
  (YsVec3 &aip,int np,const YsVec3 p[],YSBOOL strictConvexityCheck)
{
	if(YsCheckConvex3(np,p,strictConvexityCheck)==YSTRUE)
	{
		int i;
		YsVec3 cg;
		cg.Set(0.0,0.0,0.0);
		for(i=0; i<np; i++)
		{
			cg=cg+p[i];
		}
		cg=cg/double(np);
		aip=cg;
		return YSOK;
	}
	else
	{
		int i,j;
		for(i=0; i<np; i++)
		{
			for(j=i+1; j<np; j++)
			{
				if(YsCheckSeparatability3(np,p,i,j)==YSTRUE)
				{
					aip=(p[i]+p[j])/2.0;
					goto DONE;
				}
			}
		}
		YsGetCenterOfPolygon3(aip,np,p);  // I can do nothing at this point(T_T)
		return YSERR;
DONE:
		return YSOK;
	}
}

