/* This file is imported from "frames" */

#include "ysclass.h"



/*
  ax+by+c=0;
  px+qy+r=0;

  |a   b||x| |-c|
  |     || |=|  |
  |p   q||y| |-r|


  Inverted
   |ia  ib|     1  | q  -b|
   |      | = -----|      |
   |ic  id|   aq-bp|-p   a|


   |x| |ia  ib||-c|
   | |=|      ||  |
   |y| |ic  id||-r|

*/


/* ax+by+c=0 */
/* px+qy+r=0 */
YSRESULT YsSolveTwoLinearEquation
    (double *x,double *y,
     double a,double b,double c,
     double p,double q,double r)
{
	double inv;
	double ia,ib,ic,id;

	inv=a*q-b*p;
	if(inv==0)
	{
		return YSERR;
	}

	ia= q/inv;
	ib=-b/inv;
	ic=-p/inv;
	id= a/inv;

	*x=-c*ia-r*ib;
	*y=-c*ic-r*id;
	return YSOK;
}

static  YSRESULT YsEqToPlane(YsPlane &pln,double a,double b,double c,double d)
{
	YsVec3 org,nom;
	nom.Set(a,b,c);
	if(nom.GetSquareLength()<YsTolerance)
	{
		return YSERR;
	}
	nom.Normalize();
	if(YsAbs(a)>=YsAbs(b) && YsAbs(a)>=YsAbs(c))
	{
		org.Set(-d/a,0.0F,0.0F);
	}
	else if(YsAbs(b)>=YsAbs(c))
	{
		org.Set(0.0F,-d/b,0.0F);
	}
	else
	{
		org.Set(0.0F,0.0F,-d/c);
	}
	pln.Set(org,nom);
	return YSOK;
}

/*
  ax+by+cz+d=0 defines a plane. (x,y,z) from 3 equations can be calculated
  as a crossing point of 3 planes.
*/
YSRESULT YsSolveThreeLinearEquation
    (double *x,double *y,double *z,
     double a1,double b1,double c1,double d1,
     double a2,double b2,double c2,double d2,
     double a3,double b3,double c3,double d3)
{
	YsPlane pl1,pl2,pl3;
	YsVec3 lnOrg,lnVec;
	YsVec3 crs;

	YsEqToPlane(pl1,a1,b1,c1,d1);
	YsEqToPlane(pl2,a2,b2,c2,d2);
	YsEqToPlane(pl3,a3,b3,c3,d3);

	if((YsGetTwoPlaneCrossLine(lnOrg,lnVec,pl1,pl2)==YSOK &&
	    pl3.GetIntersection(crs,lnOrg,lnVec)==YSOK) ||
	   (YsGetTwoPlaneCrossLine(lnOrg,lnVec,pl1,pl3)==YSOK &&
	    pl2.GetIntersection(crs,lnOrg,lnVec)==YSOK) ||
	   (YsGetTwoPlaneCrossLine(lnOrg,lnVec,pl2,pl3)==YSOK &&
	    pl1.GetIntersection(crs,lnOrg,lnVec)==YSOK))
	{
		*x=crs.x();
		*y=crs.y();
		*z=crs.z();
		return YSOK;
	}
	return YSERR;
}


