#include <stdio.h>
#include <string.h>
#include <stdlib.h>


#include "ysdef.h"
#include "ysbase.h"
#include "ysclassmgr.h"
#include "ysmath.h"
#include "ysmatrix.h"
#include "ysgeometry.h"
#include "ysadvgeometry.h"
#include "ysprintf.h"


////////////////////////////////////////////////////////////

// Implementation //////////////////////////////////////////
#ifndef YSMATHFAST
YsClassHierarchy *YsVec2::ClassInfo=NULL;
char *YsVec2::ClassName="YsVec2";
#endif

#ifndef YSMATHFAST
YsVec2::~YsVec2()
{
	ClassInfo->Release();
}
#endif

YSBOOL YsVec2::IsNormalized(void) const
{
	double verify;
	verify=GetSquareLength();
	if(1.0-YsTolerance<=verify && verify<=1.0+YsTolerance)
	{
		return YSTRUE;
	}
	else
	{
		return YSFALSE;
	}
}

YSRESULT YsVec2::Normalize(void)
{
	double length;
	length=GetSquareLength();
	if(length>YsTolerance)
	{
		length=sqrt(length);
		value[0]/=length;
		value[1]/=length;
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

////////////////////////////////////////////////////////////

// Implementation //////////////////////////////////////////
#ifndef YSMATHFAST
YsClassHierarchy *YsVec3::ClassInfo=NULL;
char *YsVec3::ClassName="YsVec3";
#endif

#ifndef YSMATHFAST
YsVec3::~YsVec3()
{
	ClassInfo->Release();
}
#endif

YSBOOL YsVec3::IsNormalized(void) const
{
	double verify;
	verify=GetSquareLength();
	if(1.0-YsTolerance<=verify && verify<=1.0+YsTolerance)
	{
		return YSTRUE;
	}
	else
	{
		return YSFALSE;
	}
}

YSRESULT YsVec3::Normalize(void)
{
	double length;
	length=GetSquareLength();
	if(length>YsTolerance)
	{
		length=sqrt(length);
		value[0]/=length;
		value[1]/=length;
		value[2]/=length;
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YsVec3 YsVec3::GetArbitraryParpendicularVector(void) const
{
	YsVec3 vec;
	if(YsEqual(x(),0.0)==YSTRUE)
	{
		vec.Set(1.0,0.0,0.0);
	}
	else if(YsEqual(y(),0.0)==YSTRUE)
	{
		vec.Set(0.0,1.0,0.0);
	}
	else if(YsEqual(z(),0.0)==YSTRUE)
	{
		vec.Set(0.0,0.0,1.0);
	}
	else
	{
		double a,b,c;
		/* (a,b,c)(d,e,f) -> ad+be+cf=0                */
		/* P= ad ,Q=  be,R=  cf -> P+Q+R=0             */
		/* P=2abc,Q=-abc,R=-abc -> P+Q+R=0             */
	    /*   -> d=2bc,e=-ac,f=-ab                      */
		a=x();
		b=y();
		c=z();
		vec.Set(2.0*b*c,-a*c,-a*b);
		vec.Normalize();
	}
	return vec;
}

////////////////////////////////////////////////////////////

// Implementation //////////////////////////////////////////
#ifndef YSMATHFAST
YsClassHierarchy *YsRotation::ClassInfo=NULL;
char *YsRotation::ClassName="YsRotation";
#endif

YsRotation::YsRotation()
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();

//	axis.Set(0.0,0.0,1.0);
//	angle=0.0;
#endif
}

#ifndef YSMATHFAST
YsRotation::YsRotation(const YsRotation &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}
#endif

YsRotation::YsRotation(double x,double y,double z,double a)
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
#endif
	axis.Set(x,y,z);
	angle=a;
}

YsRotation::YsRotation(const YsVec3 &ax,double a)
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
#endif
	axis.Set(ax.x(),ax.y(),ax.z());
	angle=a;
}

#ifndef YSMATHFAST
YsRotation::~YsRotation()
{
	ClassInfo->Release();
}
#endif

YSRESULT YsRotation::RotatePositive(YsVec3 &result,const YsVec3 &start) const
{
	YsMatrix4x4 mat;
	mat.Rotate(*this);
	mat.Mul(result,start,0.0);
	return YSOK;
}

YSRESULT YsRotation::RotateNegative(YsVec3 &result,const YsVec3 &start) const
{
	YsMatrix4x4 mat;
	mat.Rotate(*this);
	mat.MulInverse(result,start,0.0);
	return YSOK;
}

////////////////////////////////////////////////////////////
// Implementation //////////////////////////////////////////
#ifndef YSMATHFAST
YsClassHierarchy *YsMatrix4x4::ClassInfo=NULL;
char *YsMatrix4x4::ClassName="YsMatrix4x4";
#endif

#ifndef YSMATHFAST
YsMatrix4x4::~YsMatrix4x4()
{
	ClassInfo->Release();
}
#endif

void YsMatrix4x4::Initialize(void)
{
	for(int x=0; x<4; x++)
	{
		for(int y=0; y<4; y++)
		{
			if(x==y)
			{
				value[y][x]=1.0;
				inverse[y][x]=1.0;
			}
			else
			{
				value[y][x]=0.0;
				inverse[y][x]=0.0;
			}
		}
	}
	inverseIsEffective=YSTRUE;
}

//             A B C D
//             E F G H
//             I J K L
//             M N O P
//
// a b c d
// e f g h
// i j k l
// m n o p

static void MulMatrix
    (double result[4][4],const double a[4][4],const double b[4][4])
{
	int x,y,i;
	double tmp[4][4],k;
	for(y=0; y<4; y++)
	{
		for(x=0; x<4; x++)
		{
			k=0.0;
			for(i=0; i<4; i++)
			{
				k+=a[y][i]*b[i][x];
			}
			tmp[y][x]=k;
		}
	}

	for(y=0; y<4; y++)
	{
		for(x=0; x<4; x++)
		{
			result[y][x]=tmp[y][x];
		}
	}
}

static void MulMatrix(double result[4],const double a[4][4],const double v[4])
{
	int i,y;
	double tmp[4],k;

	for(y=0; y<4; y++)
	{
		k=0.0;
		for(i=0; i<4; i++)
		{
			k+=a[y][i]*v[i];
		}
		tmp[y]=k;
	}

	for(y=0; y<4; y++)
	{
		result[y]=tmp[y];
	}
}

static void IniMatrix(double mat[4][4])
{
	for(int x=0; x<4; x++)
	{
		for(int y=0; y<4; y++)
		{
			if(x==y)
			{
				mat[y][x]=1.0;
			}
			else
			{
				mat[y][x]=0.0;
			}
		}
	}
}

void YsMatrix4x4::RotateZY(double a)
{
	// ZY Plane
	double s,c,mat[4][4];
	s=sin(a);
	c=cos(a);

	// Z'=  c Z     -s Y
	// Y'=  s Z      c Y

    // X'=  1 X      0 Y      0 Z
	// Y'=  0 X      c Y      s Z
	// Z'=  0 X     -s Y      c Z

	IniMatrix(mat);
	mat[1][1]= c;   mat[1][2]= s;
	mat[2][1]=-s;   mat[2][2]= c;
	MulMatrix(value,value,mat);

	IniMatrix(mat);
	mat[1][1]= c;   mat[1][2]=-s;
	mat[2][1]= s;   mat[2][2]= c;
	MulMatrix(inverse,mat,inverse);
}

void YsMatrix4x4::RotateXZ(double a)
{
	// XZ Plane
	double s,c,mat[4][4];
	s=sin(a);
	c=cos(a);

	// X'=  c X     -s Z
	// Z'=  s X      c Z

    // X'=  c X      0 Y     -s Z
    // Y'=  0 X      1 Y      0 Z
    // Z'=  s X      0 Y      c Z

	IniMatrix(mat);
	mat[0][0]= c;  mat[0][2]=-s;
	mat[2][0]= s;  mat[2][2]= c;
	MulMatrix(value,value,mat);

	IniMatrix(mat);
	mat[0][0]= c;  mat[0][2]= s;
	mat[2][0]=-s;  mat[2][2]= c;
	MulMatrix(inverse,mat,inverse);
}

void YsMatrix4x4::RotateXY(double a)
{
	// XY Plane
	double s,c,mat[4][4];
	s=sin(a);
	c=cos(a);

	// X'=  c X     -s Y
	// Y'=  s X      c Y

    // X'=  c X     -s Y      0 Z
    // Y'=  s X      c Y      0 Z
    // Z'=  0 X      0 Y      1 Z

	IniMatrix(mat);
	mat[0][0]= c;  mat[0][1]=-s;
	mat[1][0]= s;  mat[1][1]= c;
	MulMatrix(value,value,mat);

	IniMatrix(mat);
	mat[0][0]= c;  mat[0][1]= s;
	mat[1][0]=-s;  mat[1][1]= c;
	MulMatrix(inverse,mat,inverse);
}

#define atan3(a,b) atan2((b),(a))

void YsMatrix4x4::Rotate(double x,double y,double z,double a)
{
	double l;
	l=x*x+y*y+z*z;
	if(l==0.0)
	{
		YsPrintf::def->Printf("YsMatrix4x4::Rotate\n"
		    "  Zero Vector is specified for Rotation Axis\n");
		return;
	}

	if(l<=0.99F || 1.01F<=l)
	{
		l=sqrt(l);
		x/=l;
		y/=l;
		z/=l;
	}

	double hdg,pch;
	hdg=((x!=0.0 || z!=0.0) ? atan3(z,-x) : 0.0);
	pch=asin(y);

	RotateXZ(hdg);
	RotateZY(pch);
	RotateXY(a);
	RotateZY(-pch);
	RotateXZ(-hdg);   //
}

void YsMatrix4x4::Translate(double x,double y,double z)
{
	//
	double mat[4][4];
	IniMatrix(mat);
	mat[0][3]=x;
	mat[1][3]=y;
	mat[2][3]=z;
	MulMatrix(value,value,mat);

	IniMatrix(mat);
	mat[0][3]=-x;
	mat[1][3]=-y;
	mat[2][3]=-z;
	MulMatrix(inverse,mat,inverse);
}

void YsMatrix4x4::Scale(double x,double y,double z)
{
	double mat[4][4];
	IniMatrix(mat);
	mat[0][0]*=x;
	mat[1][1]*=y;
	mat[2][2]*=z;
	MulMatrix(value,value,mat);

	if(x==0.0 || y==0.0 || z==0.0)
	{
		YsPrintf::def->Printf("YsMatrix4x4::Scale()\n"
		        "  Multiplied by zero.\n"
		        "  Inverted Matrix Does Not Work Any More.\n");
		inverseIsEffective=YSFALSE;
	}
	else
	{
		IniMatrix(mat);
		mat[0][0]/=x;
		mat[1][1]/=y;
		mat[2][2]/=z;
		MulMatrix(inverse,mat,inverse);
	}
}

YSRESULT YsMatrix4x4::Invert(void)
{
	if(inverseIsEffective==YSTRUE)
	{
		double tmp[4][4];
		int x,y;

		for(x=0; x<4; x++)
		{
			for(y=0; y<4; y++)
			{
				tmp[y][x]=value[y][x];
			}
		}

		for(x=0; x<4; x++)
		{
			for(y=0; y<4; y++)
			{
				value[y][x]=inverse[y][x];
			}
		}

		for(x=0; x<4; x++)
		{
			for(y=0; y<4; y++)
			{
				inverse[y][x]=tmp[y][x];
			}
		}

		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

////////////////////////////////////////////////////////////

void YsMatrix4x4::Mul(double result[4],const double vec[4]) const
{
	MulMatrix(result,value,vec);
}

void YsMatrix4x4::MulInverse(double result[4],const double vec[4]) const
{
	MulMatrix(result,inverse,vec);
}

////////////////////////////////////////////////////////////

const YsMatrix4x4 operator*(const YsMatrix4x4 &t,const YsMatrix4x4 &u)
{
	//  T  V  = V'
	//  V' T~ = V

    //  T  U  V = V'
    //  V' U~ T~= V

    //  T' = T x U
    //  V'~= U~x T~

	YsMatrix4x4 tmp;
	MulMatrix(tmp.value  ,t.value  ,u.value);
	MulMatrix(tmp.inverse,u.inverse,t.inverse);

	return tmp;
}

const YsVec3 operator*(const YsMatrix4x4 &a,const YsVec3 &b)
{
	YsVec3 r;
	a.Mul(r,b,1.0);
	return r;
}

void YsMatrix4x4::GetArray(double mat[16]) const
{
	for(int x=0; x<4; x++)
	{
		for(int y=0; y<4; y++)
		{
			mat[x*4+y]=value[y][x];
		}
	}
}

YSRESULT YsMatrix4x4::GetInverseArray(double mat[16]) const
{
	if(inverseIsEffective==YSTRUE)
	{
		for(int x=0; x<4; x++)
		{
			for(int y=0; y<4; y++)
			{
				mat[x*4+y]=inverse[y][x];
			}
		}
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

// See document/mat2rot.jpg
YSRESULT YsMatrix4x4::GetRotation
    (double &x,double &y,double &z,double &rad) const
{
	double a,b,c;
	double d,e,f;
	double h,i,j;

	a=value[0][0];
	b=value[0][1];
	c=value[0][2];
	d=value[1][0];
	e=value[1][1];
	f=value[1][2];
	h=value[2][0];
	i=value[2][1];
	j=value[2][2];

	YsVec3 v1,v2,v3,v;
	double l1,l2,l3,l;
	v1.Set(a-1,b,c);
	v2.Set(d,e-1,f);
	v3.Set(h,i,j-1);

	l1=v1.GetSquareLength();
	l2=v2.GetSquareLength();
	l3=v3.GetSquareLength();

	if(l1>l2 && l1>l3)
	{
		v=v1;
		l=l1;
	}
	else if(l2>l3)
	{
		v=v2;
		l=l2;
	}
	else
	{
		v=v3;
		l=l3;
	}

	if(l<=YsTolerance*YsTolerance)
	{
		x=0.0;
		y=1.0;
		z=0.0;
		rad=0.0;
		return YSOK;
	}

	v.Normalize();

	YsVec3 u;
	u=v;
	Mul(u,v,0.0);
	u.Normalize();

	double inner;
	YsVec3 outer;
	inner=v*u;
	outer=v^u;
	outer.Normalize();

	rad=acos(inner);
	outer.Get(x,y,z);
	return YSOK;
}

YSRESULT YsMatrix4x4::GetRotation(YsRotation &rot) const
{
	double x,y,z,rad;
	if(GetRotation(x,y,z,rad)==YSOK)
	{
		rot.Set(x,y,z,rad);
	}
	return YSERR;
}

YSRESULT YsMatrix4x4::GetTranslation(double &x,double &y,double &z) const
{
	YsVec3 tmp;
	GetTranslation(tmp);
	x=tmp.x();
	y=tmp.y();
	z=tmp.z();
	return YSOK;
}

YSRESULT YsMatrix4x4::GetTranslation(YsVec3 &trs) const
{
	YsVec3 tmp(0.0,0.0,0.0);
	Mul(tmp,tmp,1.0);
	trs=tmp;
	return YSOK;
}

////////////////////////////////////////////////////////////

void YsMatrix4x4::Mul
    (YsVec3 &result,const YsVec3 &org,double translate) const
{
	double v[4];
	v[0]=org.x();
	v[1]=org.y();
	v[2]=org.z();
	v[3]=translate;
	Mul(v,v);
	result.Set(v[0],v[1],v[2]);
}

void YsMatrix4x4::MulInverse
    (YsVec3 &result,const YsVec3 &org,double translate) const
{
	double v[4];
	v[0]=org.x();
	v[1]=org.y();
	v[2]=org.z();
	v[3]=translate;
	MulInverse(v,v);
	result.Set(v[0],v[1],v[2]);
}

void YsMatrix4x4::Rotate(const YsRotation &rot)
{
	YsVec3 v;
	double a;
	rot.Get(v,a);
	Rotate(v.x(),v.y(),v.z(),a);
}

void YsMatrix4x4::Rotate(const YsAtt3 &att)
{
	(*this)=(*this)*att.GetMatrix();
}

void YsMatrix4x4::Translate(const YsVec3 &vec)
{
	Translate(vec.x(),vec.y(),vec.z());
}

void YsMatrix4x4::Scale(const YsVec3 &sca)
{
	Scale(sca.x(),sca.y(),sca.z());
}

////////////////////////////////////////////////////////////
// Implementation //////////////////////////////////////////
#ifndef YSMATHFAST
YsClassHierarchy *YsAtt3::ClassInfo=NULL;
char *YsAtt3::ClassName="YsAtt3";
#endif

YsAtt3::YsAtt3()
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
//	hdg=0.0;
//	pch=0.0;
//	bnk=0.0;
//	mat.Initialize();
#endif
}

YsAtt3::YsAtt3(double h,double p,double b)
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
#endif
	Set(h,p,b);
}

#ifndef YSMATHFAST
YsAtt3::YsAtt3(const YsAtt3 &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}
#endif

#ifndef YSMATHFAST
YsAtt3::~YsAtt3()
{
	ClassInfo->Release();
}
#endif

void YsAtt3::Set(double h,double p,double b)
{
	hdg=h;
	pch=p;
	bnk=b;
	RefreshMatrix();
}

void YsAtt3::SetH(double h)
{
	hdg=h;
	RefreshMatrix();
}

void YsAtt3::SetP(double p)
{
	pch=p;
	RefreshMatrix();
}

void YsAtt3::SetB(double b)
{
	bnk=b;
	RefreshMatrix();
}

YsVec3 YsAtt3::GetForwardVector(void) const
{
	YsVec3 vec;
	switch(YsCoordSysModel)
	{
	case YSBLUEIMPULSE:
		vec.Set(0.0,0.0,1.0);
		break;
	case YSOPENGL:
		vec.Set(0.0,0.0,-1.0);
		break;
	}
	mat.Mul(vec,vec,0.0);
	return vec;
}

YsVec3 YsAtt3::GetUpVector(void) const
{
	YsVec3 vec;
	switch(YsCoordSysModel)
	{
	case YSBLUEIMPULSE:
		vec.Set(0.0,1.0,0.0);
		break;
	case YSOPENGL:
		vec.Set(0.0,1.0,0.0);
		break;
	}
	mat.Mul(vec,vec,0.0);
	return vec;
}

YSRESULT YsAtt3::SetTwoVector(const YsVec3 &fwd,const YsVec3 &up)
{
	SetForwardVector(fwd);
	SetUpVector(up);
	return YSOK;
}

YSRESULT YsAtt3::SetForwardVector(YsVec3 ev)
{
	if(ev.IsNormalized()!=YSTRUE)
	{
		ev.Normalize();
	}

	switch(YsCoordSysModel)
	{
	case YSBLUEIMPULSE:
		hdg=atan2(-ev.x(),ev.z());
		pch=asin(ev.y());
		break;
	case YSOPENGL:
		hdg=atan2(-ev.x(),-ev.z());
		pch=asin(ev.y());
		break;
	}
	bnk=0.0;
	RefreshMatrix();

	return YSOK;
}

YSRESULT YsAtt3::SetUpVector(YsVec3 uv)
{
	if(uv.IsNormalized()!=YSTRUE)
	{
		uv.Normalize();
	}

	SetB(0.0);

	mat.MulInverse(uv,uv,0.0);

	double dBnk=0.0;
	switch(YsCoordSysModel)
	{
	case YSBLUEIMPULSE:
		dBnk=atan2(-uv.x(),uv.y());
		break;
	case YSOPENGL:
		dBnk=atan2(-uv.x(),uv.y());
		break;
	}
	SetB(dBnk);

	return YSOK;
}

void YsAtt3::NoseUp(double d)
{
	YsVec3 fwd,up;
	switch(YsCoordSysModel)
	{
	case YSBLUEIMPULSE:
		fwd.Set(0.0,sin(d), cos(d));
		up.Set (0.0,cos(d),-sin(d));
		break;
	case YSOPENGL:
		fwd.Set(0.0,sin(d),-cos(d));
		up.Set (0.0,cos(d), sin(d));
		break;
	}
	mat.Mul(fwd,fwd,0.0);
	mat.Mul(up,up,0.0);
	SetTwoVector(fwd,up);
}

void YsAtt3::YawLeft(double d)
{
	YsVec3 fwd,up;
	switch(YsCoordSysModel)
	{
	case YSBLUEIMPULSE:
		fwd.Set(-sin(d),0.0, cos(d));
		up.Set (0.0,1.0,0.0);
		break;
	case YSOPENGL:
		fwd.Set(-sin(d),0.0,-cos(d));
		up.Set (0.0,1.0,0.0);
		break;
	}
	mat.Mul(fwd,fwd,0.0);
	mat.Mul(up,up,0.0);
	SetTwoVector(fwd,up);
}

void YsAtt3::RefreshMatrix(void)
{
	mat.Initialize();
	switch(YsCoordSysModel)
	{
	case YSBLUEIMPULSE:
		mat.RotateXZ(hdg);
		mat.RotateZY(pch);
		mat.RotateXY(bnk);
		break;
	case YSOPENGL:
		mat.RotateXZ(-hdg);
		mat.RotateZY(-pch);
		mat.RotateXY(bnk);
		break;
	}
}

const YsMatrix4x4 &YsAtt3::GetMatrix(void) const
{
	return mat;
}

////////////////////////////////////////////////////////////
// Implementation //////////////////////////////////////////
#ifndef YSMATHFAST
YsClassHierarchy *YsPlane::ClassInfo=NULL;
char *YsPlane::ClassName="YsPlane";
#endif

YsPlane::YsPlane()
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();

//	nom.Set(0.0,0.0,-1.0);  // Defaulting to XY plane
#endif
}

#ifndef YSMATHFAST
YsPlane::YsPlane(const YsPlane &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}
#endif

YsPlane::YsPlane(const YsVec3 &o,const YsVec3 &n)
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
#endif
	Set(o,n);
}

#ifndef YSMATHFAST
YsPlane::~YsPlane()
{
	ClassInfo->Release();
}
#endif

YSBOOL YsPlane::CheckOnPlane(const YsVec3 &pos) const
{
	YsVec3 vec;
	vec=pos-org;
	vec.Normalize();
	return YsZero(vec*nom);
}

YSRESULT YsPlane::GetIntersection
    (YsVec3 &pos,const YsVec3 &lnOrg,const YsVec3 &lnVec) const
{
	double fa,fb;

	fa=-nom.x()*(lnOrg.x()-org.x())
	   -nom.y()*(lnOrg.y()-org.y())
	   -nom.z()*(lnOrg.z()-org.z());
	fb= nom.x()* lnVec.x()
	   +nom.y()* lnVec.y()
	   +nom.z()* lnVec.z();

	if(YsAbs(fb)<YsTolerance)
	{
		return YSERR;
	}

	double x,y,z;
	x=lnOrg.x()+lnVec.x()*fa/fb;
	y=lnOrg.y()+lnVec.y()*fa/fb;
	z=lnOrg.z()+lnVec.z()*fa/fb;
	pos.Set(x,y,z);

	return YSOK;
}

YSRESULT YsPlane::GetPenetration
    (YsVec3 &crs,const YsVec3 &p1,const YsVec3 &p2) const
{
	if(((p1-org)*nom)*((p2-org)*nom)<YsTolerance &&
	   GetIntersection(crs,p1,p2-p1)==YSOK &&
	   (p1-crs)*(p2-crs)<0.0)
	{
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsPlane::MakeBestFitPlane(int np,const YsVec3 p[])
{
	if(np>=3 && YsGetAverageNormalVector(nom,np,p)==YSOK)
	{
		if(nom.GetSquareLength()>=YsTolerance)
		{
			org=p[0];
			return YSOK;
		}
		YsPrintf::def->Printf(
		    "YsPlane::MakeBestFitPlane\n"
		    "Detected Bad Polygon\n");
	}
	return YSERR;
}

YSRESULT YsPlane::GetNearestPoint(YsVec3 &np,const YsVec3 &from) const
{
	YsVec3 ref,backVec;

	if(YsGetNearestPointOnLine3(ref,org,org+nom,from)==YSOK)
	{
		np=from+(org-ref);
		return YSERR;
	}
	return YSERR;
}
