#ifndef YSGEOMETRY_IS_INCLUDED
#define YSGEOMETRY_IS_INCLUDED
/* { */

#include <math.h>

#define YSMATHFAST  // This line disables Memory Leak Tracking

////////////////////////////////////////////////////////////


// Declaration /////////////////////////////////////////////
class YsVec2
#ifndef YSMATHFAST
: public YsMath
#endif
{
public:
#ifndef YSMATHFAST
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsVec2::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsVec2::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	inline YsVec2(const YsVec2 &from);
	virtual ~YsVec2();
#endif
	inline YsVec2();
	inline YsVec2(double x,double y);

	inline double x(void) const
	{
		return value[0];
	}
	inline double y(void) const
	{
		return value[1];
	}
	inline const double *GetValue(void) const
	{
		return value;
	}
	inline void Get(double &x,double &y) const
	{
		x=value[0];
		y=value[1];
	}
	inline double GetSquareLength(void) const
	{
		return value[0]*value[0]+value[1]*value[1];
	}
	inline double GetLength(void) const
	{
		return (double)sqrt(GetSquareLength());
	}
	YSBOOL IsNormalized(void) const;
	YSRESULT Normalize(void);
	inline void Set(double x,double y)
	{
		value[0]=x;
		value[1]=y;
	}
	inline void SetX(double x)
	{
		value[0]=x;
	}
	inline void SetY(double y)
	{
		value[1]=y;
	}
	inline void Set(double v[2])
	{
		value[0]=v[0];
		value[1]=v[1];
	}
	inline void Rotate(double ang);

	inline YsVec2 &operator+=(const YsVec2 &a);
	inline YsVec2 &operator-=(const YsVec2 &a);
	inline YsVec2 &operator/=(double a);
	inline YsVec2 &operator*=(double a);

protected:
	double value[2];
};

inline YsVec2::YsVec2()
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
//	value[0]=0.0;
//	value[1]=0.0;
#endif
}

#ifndef YSMATHFAST
inline YsVec2::YsVec2(const YsVec2 &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}
#endif

inline YsVec2::YsVec2(double x,double y)
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
#endif
	value[0]=x;
	value[1]=y;
}

inline void YsVec2::Rotate(double ang)
{
	double c,s,xx,yy;
	c=cos(ang);
	s=sin(ang);
	xx=x()*c-y()*s;
	yy=x()*s+y()*c;
	Set(xx,yy);
}

inline void YsVec2Plus(YsVec2 &result,const YsVec2 &a,const YsVec2 &b)
{
	result.Set(a.x()+b.x(),a.y()+b.y());
}

inline void YsVec2Minus(YsVec2 &result,YsVec2 const &a,const YsVec2 &b)
{
	result.Set(a.x()-b.x(),a.y()-b.y());
}

inline void YsVec2Mul(YsVec2 &result,const YsVec2 &a,double b)
{
	result.Set(a.x()*b,a.y()*b);
}

inline void YsVec2Div(YsVec2 &result,const YsVec2 &a,double b)
{
	result.Set(a.x()/b,a.y()/b);
}

inline YsVec2 operator-(const YsVec2 &a)
{
	YsVec2 b;
	b.Set(-a.x(),-a.y());
	return b;
}

inline YsVec2 operator-(const YsVec2 &a,const YsVec2 &b)
{
	YsVec2 c;
	YsVec2Minus(c,a,b);
	return c;
}

inline YsVec2 operator+(const YsVec2 &a,const YsVec2 &b)
{
	YsVec2 c;
	YsVec2Plus(c,a,b);
	return c;
}

inline YsVec2 operator/(const YsVec2 &a,double b)
{
	YsVec2 c;
	YsVec2Div(c,a,b);
	return c;
}

inline YsVec2 operator*(const YsVec2 &a,double b)
{
	YsVec2 c;
	YsVec2Mul(c,a,b);
	return c;
}

inline YsVec2 operator*(double b,const YsVec2 &a)
{
	YsVec2 c;
	YsVec2Mul(c,a,b);
	return c;
}

inline double operator*(const YsVec2 &a,const YsVec2 &b)
{
	return a.x()*b.x()+a.y()*b.y();
}

inline double operator^(const YsVec2 &a,const YsVec2 &b)
{
	// x1 y1
	// x2 y2
	double z;
	z=a.x()*b.y()-a.y()*b.x();
	return z;
}

inline int operator==(const YsVec2 &a,const YsVec2 &b)
{
	YsVec2 dif;
	dif=a-b;
	return (dif.GetSquareLength()<YsTolerance);
}

inline int operator!=(const YsVec2 &a,const YsVec2 &b)
{
	YsVec2 dif;
	dif=a-b;
	return (dif.GetSquareLength()>=YsTolerance);
}

inline YsVec2 &YsVec2::operator+=(const YsVec2 &a)
{
	value[0]+=a.value[0];
	value[1]+=a.value[1];
	return *this;
}

inline YsVec2 &YsVec2::operator-=(const YsVec2 &a)
{
	value[0]-=a.value[0];
	value[1]-=a.value[1];
	return *this;
}

inline YsVec2 &YsVec2::operator/=(double a)
{
	value[0]/=a;
	value[1]/=a;
	return *this;
}

inline YsVec2 &YsVec2::operator*=(double a)
{
	value[0]*=a;
	value[1]*=a;
	return *this;
}

////////////////////////////////////////////////////////////

// Declaration /////////////////////////////////////////////
class YsVec3
#ifndef YSMATHFAST
: public YsMath
#endif
{
public:
#ifndef YSMATHFAST
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsVec3::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsVec3::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	virtual ~YsVec3();
	inline YsVec3(const YsVec3 &from);
#endif
	inline YsVec3();
	inline YsVec3(double x,double y,double z);

	inline double x(void) const
	{
		return value[0];
	}
	inline double y(void) const
	{
		return value[1];
	}
	inline double z(void) const
	{
		return value[2];
	}
	inline const double *GetValue(void) const
	{
		return value;
	}
	inline void Set(double x,double y,double z)
	{
		value[0]=x;
		value[1]=y;
		value[2]=z;
	}
	inline void SetX(double x)
	{
		value[0]=x;
	}
	inline void SetY(double y)
	{
		value[1]=y;
	}
	inline void SetZ(double z)
	{
		value[2]=z;
	}
	inline void Set(double v[3])
	{
		value[0]=v[0];
		value[1]=v[1];
		value[2]=v[2];
	}
	inline void Get(double &x,double &y,double &z) const
	{
		x=value[0];
		y=value[1];
		z=value[2];
	}
	inline double GetSquareLength(void) const
	{
		return value[0]*value[0]+value[1]*value[1]+value[2]*value[2];
	}
	inline double GetLength(void) const
	{
		return (double)sqrt(GetSquareLength());
	}
	YSBOOL IsNormalized(void) const;
	YSRESULT Normalize(void);

	YsVec3 GetArbitraryParpendicularVector(void) const;

	inline void RotateXY(double ang);
	inline void RotateXZ(double ang);
	inline void RotateYZ(double ang);

	inline YsVec3 &operator+=(const YsVec3 &a);
	inline YsVec3 &operator-=(const YsVec3 &a);
	inline YsVec3 &operator/=(double a);
	inline YsVec3 &operator*=(double a);

protected:
	double value[3];
};

inline YsVec3::YsVec3()
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();

//	value[0]=0.0;
//	value[1]=0.0;
//	value[2]=0.0;
#endif
}

#ifndef YSMATHFAST
inline YsVec3::YsVec3(const YsVec3 &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}
#endif

inline YsVec3::YsVec3(double x,double y,double z)
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
#endif
	value[0]=x;
	value[1]=y;
	value[2]=z;
}

inline void YsVec3::RotateXY(double ang)
{
	double c,s,xx,yy;
	c=cos(ang);
	s=sin(ang);
	xx=x()*c-y()*s;
	yy=x()*s+y()*c;
	SetX(xx);
	SetY(yy);
}

inline void YsVec3::RotateXZ(double ang)
{
	double c,s,xx,zz;
	c=cos(ang);
	s=sin(ang);
	xx=x()*c-z()*s;
	zz=x()*s+z()*c;
	SetX(xx);
	SetZ(zz);
}

inline void YsVec3::RotateYZ(double ang)
{
	double c,s,yy,zz;
	c=cos(ang);
	s=sin(ang);
	yy=y()*c-z()*s;
	zz=y()*s+z()*c;
	SetY(yy);
	SetZ(zz);
}

inline void YsVec3Plus(YsVec3 &result,const YsVec3 &a,const YsVec3 &b)
{
	result.Set(a.x()+b.x(),a.y()+b.y(),a.z()+b.z());
}

inline void YsVec3Minus(YsVec3 &result,const YsVec3 &a,const YsVec3 &b)
{
	result.Set(a.x()-b.x(),a.y()-b.y(),a.z()-b.z());
}

inline void YsVec3Mul(YsVec3 &result,const YsVec3 &a,double b)
{
	result.Set(a.x()*b,a.y()*b,a.z()*b);
}

inline void YsVec3Div(YsVec3 &result,const YsVec3 &a,double b)
{
	result.Set(a.x()/b,a.y()/b,a.z()/b);
}

inline void YsVec3Outer(YsVec3 &result,const YsVec3 &a,const YsVec3 &b)
{
	// y1 z1 x1 y1
	// y2 z2 x2 y2
	double x,y,z;

	x=a.y()*b.z()-a.z()*b.y();
	y=a.z()*b.x()-a.x()*b.z();
	z=a.x()*b.y()-a.y()*b.x();
	result.Set(x,y,z);
}

inline YsVec3 operator-(const YsVec3 &a)
{
	YsVec3 b;
	b.Set(-a.x(),-a.y(),-a.z());
	return b;
}

inline YsVec3 operator-(const YsVec3 &a,const YsVec3 &b)
{
	YsVec3 c;
	YsVec3Minus(c,a,b);
	return c;
}

inline YsVec3 operator+(const YsVec3 &a,const YsVec3 &b)
{
	YsVec3 c;
	YsVec3Plus(c,a,b);
	return c;
}

inline YsVec3 operator/(const YsVec3 &a,double b)
{
	YsVec3 c;
	YsVec3Div(c,a,b);
	return c;
}

inline YsVec3 operator*(const YsVec3 &a,double b)
{
	YsVec3 c;
	YsVec3Mul(c,a,b);
	return c;
}

inline YsVec3 operator*(double b,const YsVec3 &a)
{
	YsVec3 c;
	YsVec3Mul(c,a,b);
	return c;
}

inline double operator*(const YsVec3 &a,const YsVec3 &b)
{
	return a.x()*b.x()+a.y()*b.y()+a.z()*b.z();
}

inline YsVec3 operator^(const YsVec3 &a,const YsVec3 &b)
{
	YsVec3 r;
	YsVec3Outer(r,a,b);
	return r;
}

inline int operator==(const YsVec3 &a,const YsVec3 &b)
{
	YsVec3 dif;
	dif=a-b;
	return (dif.GetSquareLength()<YsTolerance);
}

inline int operator!=(const YsVec3 &a,const YsVec3 &b)
{
	YsVec3 dif;
	dif=a-b;
	return (dif.GetSquareLength()>=YsTolerance);
}

inline YsVec3 &YsVec3::operator+=(const YsVec3 &a)
{
	value[0]+=a.value[0];
	value[1]+=a.value[1];
	value[2]+=a.value[2];
	return *this;
}

inline YsVec3 &YsVec3::operator-=(const YsVec3 &a)
{
	value[0]-=a.value[0];
	value[1]-=a.value[1];
	value[2]-=a.value[2];
	return *this;
}

inline YsVec3 &YsVec3::operator/=(double a)
{
	value[0]/=a;
	value[1]/=a;
	value[2]/=a;
	return *this;
}

inline YsVec3 &YsVec3::operator*=(double a)
{
	value[0]*=a;
	value[1]*=a;
	value[2]*=a;
	return *this;
}

////////////////////////////////////////////////////////////
// Declaration /////////////////////////////////////////////
class YsRotation
#ifndef YSMATHFAST
: public YsMath
#endif
{
public:
#ifndef YSMATHFAST
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsRotation::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsRotation::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	virtual ~YsRotation();
	YsRotation(const YsRotation &from);
#endif
	YsRotation();
	YsRotation(double x,double y,double z,double a);
	YsRotation(const YsVec3 &ax,double a);

	inline void Set(double x,double y,double z,double a)
	{
		axis.Set(x,y,z);
		angle=a;
	}
	inline void Get(double &x,double &y,double &z,double &a) const
	{
		x=axis.x();
		y=axis.y();
		z=axis.z();
		a=angle;
	}
	inline void Set(const YsVec3 &vec,double a)
	{
		axis=vec;
		angle=a;
	}
	inline void Get(YsVec3 &vec,double &a) const
	{
		vec=axis;
		a=angle;
	}
	inline const YsVec3 &GetAxis(void) const
	{
		return axis;
	}
	inline double GetAngle(void) const
	{
		return angle;
	}

	inline void SetAxis(const YsVec3 &axs)
	{
		axis=axs;
	}

	inline void SetAngle(double a)
	{
		angle=a;
	}

	YSRESULT RotatePositive(YsVec3 &result,const YsVec3 &start) const;
	YSRESULT RotateNegative(YsVec3 &result,const YsVec3 &start) const;
protected:
	double angle;
	YsVec3 axis;
};

////////////////////////////////////////////////////////////
// Declaration /////////////////////////////////////////////
class YsMatrix4x4
#ifndef YSMATHFAST
: public YsMath
#endif
{
public:
#ifndef YSMATHFAST
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsMatrix4x4::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsMatrix4x4::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	virtual ~YsMatrix4x4();
	inline YsMatrix4x4(const YsMatrix4x4 &from);
#endif
	inline YsMatrix4x4();

	friend const YsMatrix4x4 operator*(const YsMatrix4x4 &a,const YsMatrix4x4 &b);
	friend const YsVec3 operator*(const YsMatrix4x4 &a,const YsVec3 &b);

	inline double v(int r,int c) const
	{
		if(1<=r && r<=4 && 1<=c && c<=4)
		{
			return value[r-1][c-1];
		}
		else
		{
			return 0.0;
		}
	}

	void Initialize(void);
		//Set Identity Matrix
	void RotateZY(double a);
		//Rotate in Z-Y plane
	void RotateXZ(double a);
		//Rotate in X-Z plane
	void RotateXY(double a);
		//Rotate in X-Y plane
	void Rotate(double x,double y,double z,double a);
	void Rotate(const YsRotation &rot);
	void Rotate(const class YsAtt3 &att);
	void Translate(double x,double y,double z);
	void Translate(const YsVec3 &vec);
	void Scale(double x,double y,double z);
	void Scale(const YsVec3 &sca);

	void Mul(double result[4],const double vec[4]) const;
	void Mul(YsVec3 &result,const YsVec3 &org,double translate) const;
	void MulInverse(double result[4],const double vec[4]) const;
	void MulInverse(YsVec3 &result,const YsVec3 &org,double translate) const;

	YSRESULT Invert(void);

	void GetArray(double buf[16]) const;
	YSRESULT GetInverseArray(double buf[16]) const;

	YSRESULT GetRotation(double &x,double &y,double &z,double &rad) const;
	YSRESULT GetRotation(YsRotation &rot) const;
	YSRESULT GetTranslation(double &x,double &y,double &z) const;
	YSRESULT GetTranslation(YsVec3 &trs) const;

protected:
	double value[4][4];
	double inverse[4][4];
	YSBOOL inverseIsEffective;
};

inline YsMatrix4x4::YsMatrix4x4()
{
#ifndef YSMATHFAST
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
#endif
	Initialize();
}

#ifndef YSMATHFAST
inline YsMatrix4x4::YsMatrix4x4(const YsMatrix4x4 &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}
#endif

const YsMatrix4x4 operator*(const YsMatrix4x4 &a,const YsMatrix4x4 &b);
const YsVec3 operator*(const YsMatrix4x4 &a,const YsVec3 &b);


////////////////////////////////////////////////////////////
// Declaration /////////////////////////////////////////////
class YsAtt3
#ifndef YSMATHFAST
: public YsMath
#endif
{
public:
#ifndef YSMATHFAST
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsAtt3::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsAtt3::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsAtt3(const YsAtt3 &from);
	virtual ~YsAtt3();
#endif
	YsAtt3();
	YsAtt3(double h,double p,double b);

	inline double h(void) const
	{
		return hdg;
	}
	inline double p(void) const
	{
		return pch;
	}
	inline double b(void) const
	{
		return bnk;
	}
	void Set(double h,double p,double b);
	void SetH(double h);
	void SetP(double p);
	void SetB(double b);
	YsVec3 GetForwardVector(void) const;
	YsVec3 GetUpVector(void) const;
	YSRESULT SetTwoVector(const YsVec3 &fwd,const YsVec3 &up);
	YSRESULT SetForwardVector(YsVec3 vec);
	YSRESULT SetUpVector(YsVec3 vec);

	void NoseUp(double d);
	void YawLeft(double d);

	const YsMatrix4x4 &GetMatrix(void) const;

protected:
	double hdg,pch,bnk;
	YsMatrix4x4 mat;

	void RefreshMatrix(void);
};

////////////////////////////////////////////////////////////
// Declaration /////////////////////////////////////////////
class YsPlane
#ifndef YSMATHFAST
: public YsMath
#endif
{
public:
#ifndef YSMATHFAST
	static YsClassHierarchy *ClassInfo;
	static char *ClassName;
	virtual char *WhatItIs(void){return YsPlane::ClassName;}
	virtual YsClassHierarchy *Hierarchy(void){return YsPlane::ClassInfo;}
	static unsigned GetLeak(void){return ClassInfo->GetLeak();}
	YsPlane(const YsPlane &from);
	virtual ~YsPlane();
#endif
	YsPlane();
	YsPlane(const YsVec3 &o,const YsVec3 &n);

	inline void Set(const YsVec3 &o,const YsVec3 &n)
	{
		org=o;
		nom=n;
	}
	inline void SetOrigin(const YsVec3 &o)
	{
		org=o;
	}
	inline void SetNormal(const YsVec3 &n)
	{
		nom=n;
	}

	inline void GetOrigin(YsVec3 &o) const
	{
		o=org;
	}
	inline void GetNormal(YsVec3 &n) const
	{
		n=nom;
	}

	YSBOOL CheckOnPlane(const YsVec3 &pos) const;
	YSRESULT GetIntersection
	    (YsVec3 &crs,const YsVec3 &lnOrg,const YsVec3 &lnVec) const;
	YSRESULT GetPenetration
	    (YsVec3 &crs,const YsVec3 &p1,const YsVec3 &p2) const;
	YSRESULT GetNearestPoint
	    (YsVec3 &np,const YsVec3 &refp) const;
	YSRESULT MakeBestFitPlane(int np,const YsVec3 p[]);
protected:
	YsVec3 org,nom;
};

////////////////////////////////////////////////////////////

/* } */
#endif
