// NULL Check for all editing functions

#ifndef YSLIST_IS_INCLUDED
#define YSLIST_IS_INCLUDED
/* { */

#include "ysprintf.h"


// What must be updated if something is changed?
//   top,nObj,cache


template <class T> class YsList
{
public:
	YsList();
	virtual ~YsList();

	inline YsList <T> *Next(void);
	inline YsList <T> *Prev(void);
	inline YsList <T> *SeekTop(void);
	inline YsList <T> *SeekEnd(void);
	inline YsList <T> *Seek(int n);
	inline const YsList <T> *Next(void) const;
	inline const YsList <T> *Prev(void) const;
	inline const YsList <T> *SeekTop(void) const;
	inline const YsList <T> *SeekEnd(void) const;
	inline const YsList <T> *Seek(int n) const;
	YSBOOL IsTop(void) const;
	YSBOOL IsEnd(void) const;

	int GetNumObject(void) const;
		// {a0,a1,a2,...,aN)}
		//   return N+1;

	int GetPositionInList(void) const;
		// {a0,a1,a2,...,(this=an),...,(aN)}
		//   return n;
		// If called by NULL pointer,
		//   return -1;

	YsList <T> *Append(YsList <T> *follow);
		// {(this==a0),(a1),...,(aN)}
		//    +
		// {(follow==b0),(b1),...,(bN)}
		//    to
		// return {(a0),(a1),...,(aN),(b0),(b1),...,(bN)}

	YsList <T> *DeleteFromList(void);
		// {(a0),(a1),(a2),...,(a n-1),(an=this),(a n+1),...,(aN)}
		//   to
		// return {(a0),(a1),(a2),...,(a n-1),(a n+1),...,(aN)}

	YsList <T> *Delete(int n);
		// {(this=a0),(a1),(a2),...,(a n-1),(an),(a n+1),...,(aN)}
		//    to
		// return {(a0),(a1),(a2),...,(a n-1),(a n+1),...,(aN)}

	YSRESULT DeleteList(void);
		//  Delete entire node from memory.

	YsList <T> *Insert(int n,YsList <T> *toInsert);
		// {(this=a0),(a1),...,(a n-1),(an),...,(aN)}
		// {(toInsert=b0),(b1),(b2),....,(bN)}
		//    to
		// return {(a0),(a1),...,(a n-1),(b0),...,(bN),(an),...,(aN)}
		//
		// if this==NULL
		//   return toInsert;

	YsList <T> *InsertOnTheLeft(YsList <T> *toInsert);
		// {(a0),(a1),...,(a n-1),(an=this),(a n+1),...,(aN)}
		//   to
		// {(a0),(a1),...,(a n-1),{toInsert},(an=this),(a n+1),...,(aN)}

	YsList <T> *InsertOnTheRight(YsList <T> *toInsert);
		// {(a0),(a1),...,(a n-1),(an=this),(a n+1),...,(aN)}
		//   to
		// {(a0),(a1),...,(a n-1),(an=this),{toInsert},(a n+1),...,(aN)}

	YsList <T> *Split(YsList <T> *splitPoint);
		// {(a0),(a1),...,(a n-1),(an=splitPoint),...,(aN)}
		//   to
		// return {(a0),(a1),...,(a n-1)}
		//
		// {(an=splitPoint),...,(aN)} can be used as an independent list.

	YsList <T> *Duplicate(void) const;
		// return (duplicated list)

	void Test(void) const;

	void Encache(void) const;
	void Decache(void) const;

	T dat;

protected:
	YsList <T> *prv,*nxt,*top,*btm;

	// nObj and cache is available only for top of the list
	mutable int nCache;
	mutable YsList <T> **cache;
	mutable int positionInList;

	void SetNext(YsList <T> *neoNxt);
	void SetPrev(YsList <T> *neoPrv);
};

template <class T> YsList <T>::YsList()
{
	prv=NULL;
	nxt=NULL;
	top=this;
	btm=this;

	nCache=0;
	cache=NULL;

	positionInList=-1;
}

template <class T> YsList <T>::~YsList()
{
	if(cache!=NULL)
	{
		delete [] cache;
	}
}

template <class T> inline YsList <T> *YsList <T>::Next(void)
{
	const YsList <T> *cThis;
	cThis=this;
	return (YsList <T> *)(cThis->Next());
}

template <class T> inline YsList <T> *YsList <T>::Prev(void)
{
	const YsList <T> *cThis;
	cThis=this;
	return (YsList <T> *)(cThis->Prev());
}

template <class T> inline YsList <T> *YsList <T>::SeekTop(void)
{
	const YsList <T> *cThis;
	cThis=this;
	return (YsList <T> *)(cThis->SeekTop());
}

template <class T> inline YsList <T> *YsList <T>::SeekEnd(void)
{
	const YsList <T> *cThis;
	cThis=this;
	return (YsList <T> *)(cThis->SeekEnd());
}

template <class T> inline YsList <T> *YsList <T>::Seek(int n)
{
	const YsList <T> *cThis;
	cThis=this;
	return (YsList <T> *)(cThis->Seek(n));
}

template <class T> inline const YsList <T> *YsList <T>::Next(void) const
{
	if(this!=NULL)
	{
		return nxt;
	}
	else
	{
		return NULL;
	}
}

template <class T> inline const YsList <T> *YsList <T>::Prev(void) const
{
	if(this!=NULL)
	{
		return prv;
	}
	else
	{
		return NULL;
	}
}

template <class T> inline const YsList <T> * YsList <T>::SeekTop(void) const
{
	if(this==NULL)
	{
		return NULL;
	}
	else
	{
		return top;
	}

}

template <class T> inline const YsList <T> * YsList <T>::SeekEnd(void) const
{
	if(this==NULL)
	{
		return NULL;
	}
	else
	{
		return btm;
	}
}

template <class T> inline const YsList <T> * YsList <T>::Seek(int n) const
{
	if(this==NULL || n<0)
	{
		return NULL;
	}
	else if(IsTop()==YSTRUE)
	{
		if(0<=n && n<nCache && cache!=NULL)
		{
			return cache[n];
		}
		else
		{
			const YsList <T> *seeker;
			int i;
			seeker=this;
			for(i=0; i<n && seeker!=NULL; i++)
			{
				seeker=seeker->Next();
			}
			return seeker;
		}
	}
	else
	{
		YsPrintf::def->Printf("YsList::Seek()\n"
		         "  WARNING:Seek() is called for non-top member.\n");
		return top->Seek(n);
	}
}

template <class T> YSBOOL YsList <T>::IsTop(void) const
{
	if(this!=NULL && Prev()==NULL)
	{
		return YSTRUE;
	}
	else
	{
		return YSFALSE;
	}
}

template <class T> YSBOOL YsList <T>::IsEnd(void) const
{
	if(this!=NULL && Next()==NULL)
	{
		return YSTRUE;
	}
	else
	{
		return YSFALSE;
	}
}

template <class T> int YsList<T>::GetNumObject(void) const
{
	if(this!=NULL)
	{
		if(IsTop()!=YSTRUE)
		{
			YsPrintf::def->Printf("YsList<T>::GetNumObject()\n"
			         "  WARNING:Used for non-top node.\n");
		}

		const YsList <T> *top;
		top=SeekTop();
		if(top->nCache>0 && cache!=NULL)
		{
			return nCache;
		}
		else
		{
			int i;
			const YsList <T> *seeker;
			i=0;
			for(seeker=this; seeker!=NULL; seeker=seeker->Next())
			{
				i++;
			}
			return i;
		}
	}
	return 0;
}

template <class T> int YsList <T>::GetPositionInList(void) const
{
	// {(this=a0),(a1),(a2),...,(ptr==an),...,(aN)}
	// then return n;

	if(positionInList>=0)
	{
		return positionInList;
	}

	if(this!=NULL)
	{
		int i;
		const YsList <T> *seeker;

		i=0;
		for(seeker=SeekTop(); seeker!=NULL; seeker=seeker->Next())
		{
			if(seeker==this)
			{
				return i;
			}
			i++;
		}
		YsPrintf::def->Printf("YsList <T>::GetPositionInList()\n"
		         "  WARNING:Inconsistent top-node combination.\n");
	}
	return -1;
}

template <class T> YsList <T> * YsList <T>::Append(YsList <T> *follow)
{
	// {(this==a0),(a1),...,(aN)}
	//    +
	// {(follow==b0),(b1),...,(bN)}
	//    to
	// return {(a0),(a1),...,(aN),(b0),(b1),...,(bN)}

	if(follow==NULL && this==NULL)
	{
		return NULL;
	}
	else if(follow==NULL) // && this!=NULL
	{
		if(this->IsTop()==YSTRUE)
		{
			return this;
		}
		else
		{
			YsPrintf::def->Printf("YsList <T>::Append()\n"
			         "  WARNING:Used for non-top node.\n");
			return SeekTop();
		}
	}
	else if(this==NULL) // && follow!=NULL
	{
		if(follow->IsTop()==YSTRUE)
		{
			return follow;
		}
		else
		{
			YsPrintf::def->Printf("YsList <T>::Append()\n"
			         "  WARNING:Target node is not a top node.\n");
			return follow->SeekTop();
		}
	}
	else
	{
		YsList <T> *l1,*l2;
		l1=this;
		l2=follow;
		if(this->IsTop()!=YSTRUE)
		{
			YsPrintf::def->Printf("YsList <T>::Append()\n"
			         "  WARNING:Used by a non-top node.\n");
			l1=this->SeekTop();
		}
		if(follow->IsTop()!=YSTRUE)
		{
			YsPrintf::def->Printf("YsList <T>::Append()\n"
			         "  WARNING:Target node is not a top node.\n");
			l2=follow->SeekTop();
		}

		l1->Decache();
		l2->Decache();

		if(l1==l2)
		{
			YsPrintf::def->Printf("YsList <T>::Append()\n"
			         "  Two lists are identical\n");
			return l1;
		}

		// l1End<->l2Top
		YsList <T> *l1End,*l2Top;
		l1End=l1->SeekEnd();
		l2Top=l2;
		l1End->SetNext(l2Top);
		l2Top->SetPrev(l1End);

		// Reset "top" of all l2 members
		YsList <T> *ptr;
		for(ptr=l2; ptr!=NULL; ptr=ptr->Next())
		{
			ptr->top=l1;
		}

		for(ptr=l1; ptr!=l2; ptr=ptr->Next())
		{
			ptr->btm=l2->btm;
		}

		return l1;
	}
}

template <class T> YsList <T> * YsList <T>::DeleteFromList(void)
{
	// {(a0),(a1),(a2),...,(a n-1),(an=this),(a n+1),...,(aN)}
	//   to
	// return {(a0),(a1),(a2),...,(a n-1),(a n+1),...,(aN)}

	if(this==NULL)
	{
		return NULL;
	}
	else
	{
		if(IsTop()==YSTRUE)
		{
			YsList <T> *neoTop,*seeker;
			neoTop=Next();

			neoTop->SetPrev(NULL);
			for(seeker=neoTop; seeker!=NULL; seeker=seeker->Next())
			{
				seeker->top=neoTop;
			}

			delete this;

			neoTop->Decache();

			return neoTop;
		}
		else if(IsEnd()==YSTRUE)
		{
			YsList <T> *top;
			top=SeekTop();
			if(top->cache!=NULL && top->nCache>0)
			{
				top->nCache--;
			}

			YsList <T> *neoBtm,*seeker;
			neoBtm=Prev();
			neoBtm->SetNext(NULL);
			for(seeker=neoBtm; seeker!=NULL; seeker=seeker->Prev())
			{
				seeker->btm=neoBtm;
			}

			delete this;

			return top;
		}
		else
		{
			YsList <T> *n1,*n2;
			n1=Prev();
			n2=Next();
			n1->SetNext(n2);
			n2->SetPrev(n1);

			YsList <T> *top;
			top=SeekTop();

			if(top->cache!=NULL && top->nCache>0 && positionInList>=0)
			{
				int i;
				top->nCache--;  // Do it before the for-loop below.
				for(i=positionInList; i<top->nCache; i++)
				{
					top->cache[i]=top->cache[i+1];
				}
			}

			delete this;

			return top;
		}
	}
}


template <class T> YsList <T> * YsList <T>::Delete(int n)
{
	// {(this=a0),(a1),(a2),...,(a n-1),(an),(a n+1),...,(aN)}
	//    to
	// return {(a0),(a1),(a2),...,(a n-1),(a n+1),...,(aN)}

	if(this==NULL)
	{
		return NULL;
	}
	else if(IsTop()!=YSTRUE)
	{
		YsPrintf::def->Printf("YsList <T>::Delete\n"
		         "  Used by a non-top node.\n");
		return SeekTop()->Delete(n);
	}
	else
	{
		YsList <T> *target;

		Decache();

		target=Seek(n);
		if(target!=NULL)
		{
			return target->DeleteFromList();
		}
		else
		{
			return this;
		}
	}
}

template <class T> YSRESULT YsList <T>::DeleteList(void)
{
	YsList <T> *ptr,*nxt;
	for(ptr=SeekTop(); ptr!=NULL; ptr=nxt)
	{
		nxt=ptr->Next();
		delete ptr;
	}
	return YSOK;
}

template <class T> YsList <T> * YsList <T>::Insert(int n,YsList <T> *toIns)
{
	// {(this=a0),(a1),...,(a n-1),(an),...,(aN)}
	// {(toInsert=b0),(b1),(b2),....,(bN)}
	//    to
	// return {(a0),(a1),...,(a n-1),(b0),...,(bN),(an),...,(aN)}

	if(this==NULL && toIns==NULL)
	{
		return toIns;
	}
	else if(toIns==NULL)
	{
		if(this->IsTop()!=YSTRUE)
		{
			YsPrintf::def->Printf("YsList <T>::Insert()\n"
			         "  WARNING:Used by a non-top member.\n");
		}
		return this->SeekTop();
	}
	else if(this==NULL)
	{
		if(toIns->IsTop()!=YSTRUE)
		{
			YsPrintf::def->Printf("YsList <T>::Insert()\n"
			         "  WARNING:Target is not a top member.\n");
		}
		return toIns->SeekTop();
	}
	else
	{
		if(IsTop()!=YSTRUE)
		{
			YsPrintf::def->Printf("YsList <T>::Insert\n"
			         "  WARNING:Used by non-top member.\n");
			return SeekTop()->Insert(n,toIns);
		}
		if(toIns->IsTop()!=YSTRUE)
		{
			YsPrintf::def->Printf("YsList <T>::Insert\n"
			         "  WARNING:Target is not a top member.\n");
			return Insert(n,toIns->SeekTop());
		}

		if(n<=0)
		{
			return toIns->Append(this);
		}
		else if(GetNumObject()<=n)
		{
			return this->Append(toIns);
		}
		else
		{
			YsList <T> *seeker,*n1,*n2,*i1,*i2;

			n1=Seek(n-1);
			i1=toIns->SeekTop();
			i2=toIns->SeekEnd();
			n2=Seek(n);

			n1->SetNext(i1);
			i1->SetPrev(n1);
			i2->SetNext(n2);
			n2->SetPrev(i2);

			for(seeker=i1; seeker!=n2; seeker=seeker->Next())
			{
				seeker->top=SeekTop();
				seeker->btm=SeekEnd();
			}

			toIns->Decache();
			this->Decache();
			return this;
		}
	}
}

template <class T>
YsList <T> *YsList <T>::InsertOnTheLeft(YsList <T> *toInsert)
{
	if(this==NULL)
	{
		return toInsert;
	}
	else
	{
		int n;
		YsList <T> *top;
		top=SeekTop();
		n=GetPositionInList();
		return top->Insert(n,toInsert);
	}
}

template <class T>
YsList <T> *YsList <T>::InsertOnTheRight(YsList <T> *toInsert)
{
	if(this==NULL)
	{
		return toInsert;
	}
	else
	{
		int n;
		YsList <T> *top;
		top=SeekTop();
		n=GetPositionInList();
		return top->Insert(n+1,toInsert);
	}
}

template <class T> YsList <T> *YsList <T>::Split(YsList <T> *splitPoint)
{
	if(this==NULL)
	{
		return NULL;
	}
	else
	{
		if(IsTop()!=YSTRUE)
		{
			YsPrintf::def->Printf("YsList <T>::Split()\n"
			         "  WARNING:Used by non-top member.\n");
			return SeekTop()->Split(splitPoint);
		}

		if(splitPoint==this)
		{
			// {a0,a1,...,aN}
			//   into
			// {} {a0,a1,...,aN}
			return NULL;
		}
		else if(splitPoint==NULL)
		{
			// {a0,a1,...,aN}
			//  into
			// {a0,a1,...,aN} {}
			return this;
		}
		else
		{
			int n;
			n=splitPoint->GetPositionInList();
			if(n<0)
			{
				YsPrintf::def->Printf("YsList <T>::Split()\n"
				         "  Split point is not a member of the list.\n");
				return this;
			}

			YsList <T> *l1End,*l2Top,*seeker;
			l1End=splitPoint->Prev();
			l2Top=splitPoint;

			l1End->SetNext(NULL);
			l2Top->SetPrev(NULL);

			for(seeker=this; seeker!=NULL; seeker=seeker->Next())
			{
				seeker->btm=l1End;
			}
			for(seeker=l2Top; seeker!=NULL; seeker=seeker->Next())
			{
				seeker->top=l2Top;
			}

			Decache();

			return this;
		}
	}
}

template <class T> void YsList <T>::Test(void) const
{
	int i,n;
	const YsList <T> *seeker;

	if(this==NULL)
	{
		return;
	}

	if(IsTop()!=YSTRUE)
	{
		YsPrintf::def->Printf("YsList <T>::Test\n");
		YsPrintf::def->Printf("  Used for non-top member.\n");
		SeekTop()->Test();
		return;
	}

	YsPrintf::def->Printf("--Self Diagnostics--\n");
	YsPrintf::def->Printf("YsList::Test()\n");

	if(nCache>0 && cache!=NULL)
	{
		YsPrintf::def->Printf(">>Encached\n");
	}
	else
	{
		YsPrintf::def->Printf(">>Not cached\n");
	}


	YsPrintf::def->Printf("--Check number of objects--\n");
	n=GetNumObject();
	i=0;
	for(seeker=this; seeker!=NULL; seeker=seeker->Next())
	{
		i++;
	}
	if(n==i)
	{
		YsPrintf::def->Printf(">>GetNumObj()  OK.\n");
	}
	else
	{
		YsPrintf::def->Printf(">>Number of object no match\n");
		YsPrintf::def->Printf("  GetNumObj()    :%d\n",n);
		YsPrintf::def->Printf("  Actual         :%d\n",i);
	}


	YsPrintf::def->Printf("--Check Node Pointer Matching--\n");
	i=0;
	for(seeker=this; seeker!=NULL; seeker=seeker->Next())
	{
		if(seeker==Seek(i))
		{
		//	YsPrintf::def->Printf("[%4d] OK.\n",i);
		}
		else
		{
			YsPrintf::def->Printf("[%4d] ERROR.\n",i);
		}
		i++;
	}

	YsPrintf::def->Printf("--Check Top/End Matching--\n");
	for(seeker=this; seeker!=NULL; seeker=seeker->Next())
	{
		if(seeker->SeekTop()==SeekTop() && seeker->SeekTop()==top &&
		   seeker->top==top)
		{
		//	YsPrintf::def->Printf("[%4d] OK.\n",i);
		}
		else
		{
			YsPrintf::def->Printf("[%4d] ERROR(top).\n",i);
		}

		if(seeker->SeekEnd()==SeekEnd() && seeker->SeekEnd()==btm &&
		   seeker->btm==btm)
		{
		//	YsPrintf::def->Printf("[%4d] OK.\n",i);
		}
		else
		{
			YsPrintf::def->Printf("[%4d] ERROR(top).\n",i);
		}
		i++;
	}

	YsPrintf::def->Printf("--Cache Check--\n");
	i=0;
	for(seeker=this; seeker!=NULL; seeker=seeker->Next())
	{
		if(seeker->IsTop()!=YSTRUE &&
		   (seeker->cache!=NULL || seeker->nCache>0))
		{
			YsPrintf::def->Printf("[%4d] ERROR.\n",i);
		}
		i++;
	}
}

template <class T> void YsList <T>::Encache(void) const
{
	if(this!=NULL)
	{
		Decache();

		int i,n;
		const YsList <T> *seeker;
		n=0;
		for(seeker=this; seeker!=NULL; seeker=seeker->Next())
		{
			n++;
		}

		cache=new YsList <T> *[n];
		if(cache!=NULL)
		{
			seeker=this;
			for(i=0; i<n; i++)
			{
				seeker->positionInList=i;
				cache[i]=(YsList *)seeker;
				seeker=seeker->Next();
			}
			nCache=n;
		}
	}
}

template <class T> void YsList <T>::Decache(void) const
{
	if(this!=NULL)
	{
		if(IsTop()!=YSTRUE)
		{
			YsPrintf::def->Printf("YsList::Decache()\n"
			       "  Decache() is used to non-top member.\n");
		}

		if(cache!=NULL)
		{
			const YsList <T> *seeker;
			for(seeker=this; seeker!=NULL; seeker=seeker->Next())
			{
				seeker->positionInList=-1;
			}

			delete [] cache;
		}
		nCache=0;
		cache=NULL;
	}
}

template <class T> inline void YsList <T>::SetNext(YsList <T> *neoNext)
{
	if(this!=NULL)
	{
		nxt=neoNext;
	}
}

template <class T> inline void YsList <T>::SetPrev(YsList <T> *neoPrev)
{
	if(this!=NULL)
	{
		prv=neoPrev;
	}
}

template <class T> YsList<T> *YsList <T>::Duplicate(void) const
{
	const YsList <T> *ptr;
	YsList <T> *neo,*dup;
	dup=NULL;
	for(ptr=this; ptr!=NULL; ptr=ptr->Next())
	{
		neo=new YsList <T>;
		if(neo!=NULL)
		{
			neo->dat=ptr->dat;
			dup=dup->Append(neo);
		}
		else
		{
			YsPrintf::def->Printf("YsList::Duplicate\nLowMemoryWarning\n");
			break;
		}
	}
	return dup;
}

/* } */
#endif
