#include <stdio.h>
#include <math.h>

// 02/14/1999
//   YsMatrix::~YsMatrix
//     delete dat; -> delete [] dat;

#include "ysdef.h"
#include "ysbase.h"
#include "ysclassmgr.h"
#include "ysmath.h"
#include "ysmatrix.h"



const char *PRINTFORMAT="%6.3lf";



// Implementation //////////////////////////////////////////
YsClassHierarchy *YsMatrix::ClassInfo=NULL;
char *YsMatrix::ClassName="YsMatrix";

YsMatrix::YsMatrix()
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();

	nRow=0;
	nColumn=0;
	dat=NULL;
	tmp=NULL;
}

YsMatrix::YsMatrix(const YsMatrix &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsMath::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}

YsMatrix::~YsMatrix()
{
	ClassInfo->Release();

	if(tmp!=NULL)
	{
		delete tmp;
	}
	if(dat!=NULL)
	{
		delete [] dat;
		dat=NULL;
	}
}

void YsMatrix::CreateWithoutClear(int nr,int nc)
{
	if(dat==NULL || nRow!=nr || nColumn!=nc)
	{
		if(dat!=NULL)
		{
			delete dat;
		}
		nRow=nr;
		nColumn=nc;
		dat=new double[nc*nr];
	}
}

void YsMatrix::Create(int nr,int nc)
{
	CreateWithoutClear(nr,nc);
	for(int i=0; i<nc*nr; i++)
	{
		dat[i]=0.0;
	}
}

void YsMatrix::Create(int nr,int nc,double *d)
{
	CreateWithoutClear(nr,nc);
	for(int i=0; i<nr*nc; i++)
	{
		dat[i]=d[i];
	}
}

void YsMatrix::LoadIdentity(void)
{
	int i;
	for(i=1; i<=nRow; i++)
	{
		Set(i,i,1.0);
	}
}

YSRESULT YsMatrix::GetSubMatrix(YsMatrix &sub,int r,int c) const
{
	int x,y,p,q;
	static const double modulus[2]={1.0,-1.0};

	sub.CreateWithoutClear(nRow-1,nColumn-1);
	q=1;
	for(y=1; y<=nRow; y++)
	{
		p=1;
		if(y!=r)
		{
			for(x=1; x<=nColumn; x++)
			{
				if(x!=c)
				{
					sub.Set(q,p,v(y,x));
					p++;
				}
			}
			q++;
		}
	}

	return YSOK;
}

YSRESULT YsMatrix::GetDeterminantMatrix(YsMatrix &m) const
{
	int r,c;
	double detSub;
	YsMatrix sub;
	static const double modulus[2]={1.0,-1.0};

	m.CreateWithoutClear(nRow,nColumn);
	for(r=1; r<=nRow; r++)
	{
		for(c=1; c<=nColumn; c++)
		{
			GetSubMatrix(sub,r,c);
			sub.GetDeterminant(detSub);
			m.Set(r,c,detSub*modulus[(r+c)&1]);
		}
	}
	m.Transpose();

	return YSOK;
}

static int lev=0;
YSRESULT YsMatrix::GetDeterminant(double &det) const
{
	if(nColumn!=nRow)
	{
		return YSERR;
	}

	if(nRow==2)
	{
		det=v(1,1)*v(2,2)-v(1,2)*v(2,1);
	}
	else
	{
		int i;
		double sum;
		YsMatrix m;
		GetDeterminantMatrix(m);
		sum=0.0;
		for(i=1; i<=nRow; i++)
		{
			sum+=m.v(1,i)*v(i,1);
		}
		det=sum;
	}
	return YSOK;
}

YSRESULT YsMatrix::Transpose(void)
{
	if(nRow!=nColumn)
	{
		return YSERR;
	}
	else
	{
		int r,c;
		double a,b;
		for(r=1; r<=nRow; r++)
		{
			for(c=r; c<=nColumn; c++)
			{
				a=v(r,c);
				b=v(c,r);
				Set(r,c,b);
				Set(c,r,a);
			}
		}
		return YSOK;
	}
}

//YSRESULT YsMatrix::Invert(void)
//{
//	if(nColumn!=nRow)
//	{
//		return YSERR;
//	}
//
//	int i;
//	double det;
//	YsMatrix m;
//	GetDeterminantMatrix(m);
//	det=0.0;
//	for(i=1; i<=nRow; i++)
//	{
//		det+=m.v(1,i)*v(i,1);
//	}
//
//	if(det!=0.0)
//	{
//		int r,c;
//		for(r=1; r<=nRow; r++)
//		{
//			for(c=1; c<=nColumn; c++)
//			{
//				Set(r,c,m.v(r,c)/det);
//			}
//		}
//		return YSOK;
//	}
//	else
//	{
//		return YSERR;
//	}
//}

YSRESULT YsMatrix::SwapRow(int r1,int r2)
{
	if(1<=r1 && r1<=nRow && 1<=r2 && r2<=nRow)
	{
		double a,b;
		int c;
		for(c=1; c<=nColumn; c++)
		{
			a=v(r1,c);
			b=v(r2,c);
			Set(r1,c,b);
			Set(r2,c,a);
		}
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsMatrix::MulRow(int r,double m)
{
	if(1<=r && r<=nRow)
	{
		int c;
		for(c=1; c<=nColumn; c++)
		{
			Set(r,c,v(r,c)*m);
		}
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsMatrix::DivRow(int r,double m)
{
	if(1<=r && r<=nRow)
	{
		int c;
		for(c=1; c<=nColumn; c++)
		{
			Set(r,c,v(r,c)/m);
		}
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsMatrix::Row1MinusRow2Mul(int r1,int r2,double m)
{
	if(1<=r1 && r1<=nRow && 1<=r2 && r2<=nRow)
	{
		int c;
		for(c=1; c<=nColumn; c++)
		{
			double a;
			a=v(r1,c)-v(r2,c)*m;
			Set(r1,c,a);
		}
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsMatrix::Invert(void)
{
	YsMatrix sync;

	sync.Create(nRow,nColumn);
	sync.LoadIdentity();

	int n,r;
	double a;
	for(n=1; n<=nColumn; n++)
	{
		// Sweep Out Column n
		for(r=n+1; r<=nRow; r++)
		{
			if(YsAbs(v(n,n))<YsAbs(v(r,n)))
			{
				//printf("Swap %d<->%d\n",r,n);
				SwapRow(r,n);
				sync.SwapRow(r,n);
			}
		}

		a=v(n,n);
		if(YsAbs(a)>=YsTolerance)
		{
			DivRow(n,a);
			sync.DivRow(n,a);

			//printf("MinusRow %d x %lf\n",n,a);
			for(r=1; r<=nRow; r++)
			{
				if(r!=n)
				{
					a=v(r,n);
					Row1MinusRow2Mul(r,n,a);
					sync.Row1MinusRow2Mul(r,n,a);
				}
			}
		}
		else
		{
			printf("Cannot Eliminate Column %d\n!",n);
			return YSERR;
		}

		//Print();
		//sync.Print();
  		//char du[256];
  		//printf("(1:%d/%d)>",n,nColumn);
  		//gets(du);
	}
	*this=sync;

	return YSOK;
}

void YsMatrix::Print(void) const
{
	printf("----\n");
	for(int r=1; r<=nRow; r++)
	{
		for(int c=1; c<=nColumn; c++)
		{
			printf(PRINTFORMAT,v(r,c));
		}
		printf("\n");
	}
}

const YsMatrix &YsMatrix::operator=(const YsMatrix &a)
{
	if(this!=&a)
	{
		Create(a.nRow,a.nColumn,a.dat);
	}
	return *this;
}


YsMatrix *YsMatrix::UseTemporaryMatrix(void) const
{
	if(tmp!=NULL)
	{
		return tmp;  // Direct Recycle. Good for environment?
	}
	else
	{
		tmp=new YsMatrix;
		if(tmp==NULL)
		{
			printf("YsMatrix::UseTemporaryMatrix(void)\n");
			printf("Low memory warning!!!\n");
		}
		return tmp;
	}
}

//
//int main(void)
//{
//	double mat[5*5]=
//	{
//		 0, 0, 3, 0, 8,
//		 1, 0, 1, 0, 0,
//		 0, 1, 0, 1,-1,
//		 0,-8, 0,-8,-1,
//		 0, 0, 0, 0, 1
//	};
//	double cst[5]=
//	{
//		0,0,0,0,2600
//	};
//	YsMatrix m1,m2,c,a;
//
//	m1.Create(5,5,mat);
//	m2=m1;
//	m2.InvertBySweepingOut();
//	c.Create(5,1,cst);
//	a=m2*c;
//	a.Print();
//
//	return 0;
//}
