#include "ysclass.h"

// Implementation //////////////////////////////////////////
YsClassHierarchy *YsWizard::ClassInfo=NULL;
char *YsWizard::ClassName="YsWizard";

YsWizard::YsWizard()
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsBaseClass::ClassInfo);
	}
	ClassInfo->Allocate();
}

YsWizard::YsWizard(const YsWizard &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsBaseClass::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}

YsWizard::~YsWizard()
{
	ClassInfo->Release();
}

// Implementation //////////////////////////////////////////
YsClassHierarchy *YsCollisionOfPolygon::ClassInfo=NULL;
char *YsCollisionOfPolygon::ClassName="YsCollisionOfPolygon";

YsCollisionOfPolygon::YsCollisionOfPolygon()
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsWizard::ClassInfo);
	}
	ClassInfo->Allocate();

	np1=0;
	p1=NULL;
	np2=0;
	p2=NULL;
}

YsCollisionOfPolygon::YsCollisionOfPolygon(const YsCollisionOfPolygon &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsWizard::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}

YsCollisionOfPolygon::~YsCollisionOfPolygon()
{
	ClassInfo->Release();
}

void YsCollisionOfPolygon::SetPolygon1(int nv1,const YsVec3 v1[])
{
	if(nv1>0)
	{
		np1=nv1;
		p1=v1;

		YsBoundingBoxMaker3 bbx;
		bbx.Make(nv1,v1);
		bbx.Get(p1min,p1max);

		YsVec3 nom;
		YsGetAverageNormalVector(nom,nv1,v1);
		pln1.Set(v1[0],nom);
	}
}

void YsCollisionOfPolygon::SetPolygon2(int nv2,const YsVec3 v2[])
{
	if(nv2>0)
	{
		np2=nv2;
		p2=v2;

		YsBoundingBoxMaker3 bbx;
		bbx.Make(nv2,v2);
		bbx.Get(p2min,p2max);

		YsVec3 nom;
		YsGetAverageNormalVector(nom,nv2,v2);
		pln2.Set(v2[0],nom);
	}
}

YSBOOL YsCollisionOfPolygon::CheckOnTheSamePlane(void) const
{
	YsVec3 n1,n2,o1;
	pln1.GetNormal(n1);
	pln2.GetNormal(n2);

	if((n1^n2).GetSquareLength()<=YsTolerance)
	{
		pln1.GetOrigin(o1);
		return pln2.CheckOnPlane(o1);
	}
	return YSFALSE;
}

YSBOOL YsCollisionOfPolygon::CheckCollision(void) const
{
	int i;
	const YsVec3 *v1,*v2;
	YSSIDE is;
	YsVec3 crs;

	if(YsCheckBoundingBoxCollision3(p1min,p1max,p2min,p2max)==YSTRUE)
 	{
		if(np1>=2 || np2>=3) // Polygon/LineSeg1 vs Plane2
		{
			for(i=0; i<np1; i++)
			{
				v1=&p1[i];
				v2=&p1[(i+1)%np1];
				if(pln2.GetPenetration(crs,*v1,*v2)==YSOK)
				{
					is=YsCheckInsidePolygon3(crs,np2,p2);
					if(is==YSINSIDE || is==YSBOUNDARY)
					{
						return YSTRUE;
					}
				}
			}
		}
		if(np1>=3 || np2>=2) // Polygon/LineSeg2 vs Plane1
		{
			for(i=0; i<np2; i++)
			{
				v1=&p2[i];
				v2=&p2[(i+1)%np2];
				if(pln1.GetPenetration(crs,*v1,*v2)==YSOK)
				{
					is=YsCheckInsidePolygon3(crs,np1,p1);
					if(is==YSINSIDE || is==YSBOUNDARY)
					{
						return YSTRUE;
					}
				}
			}
		}
	}
	return YSFALSE;
}

YSINTERSECTION YsCollisionOfPolygon::CheckPenetration(void) const
{
	if(YsCheckBoundingBoxCollision3(p1min,p1max,p2min,p2max)==YSTRUE)
 	{
		return YsGetPolygonPenetration(np1,p1,np2,p2);
	}
	return YSAPART;
}

YSBOOL YsCollisionOfPolygon::OneEdgeLiesOnTheOtherPlane(void) const
{
	int i;
	for(i=0; i<np1; i++)
	{
		if(pln2.CheckOnPlane(p1[i])==YSTRUE &&
		   pln2.CheckOnPlane(p1[(i+1)%np1])==YSTRUE)
		{
			return YSTRUE;
		}
	}

	for(i=0; i<np2; i++)
	{
		if(pln1.CheckOnPlane(p2[i])==YSTRUE &&
		   pln1.CheckOnPlane(p2[(i+1)%np1])==YSTRUE)
		{
			return YSTRUE;
		}
	}

	return YSFALSE;
}

// Implementation //////////////////////////////////////////
YsClassHierarchy *YsBoundingBoxMaker2::ClassInfo=NULL;
char *YsBoundingBoxMaker2::ClassName="YsBoundingBoxMaker2";

YsBoundingBoxMaker2::YsBoundingBoxMaker2()
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsWizard::ClassInfo);
	}
	ClassInfo->Allocate();
}

YsBoundingBoxMaker2::YsBoundingBoxMaker2(const YsBoundingBoxMaker2 &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsWizard::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}

YsBoundingBoxMaker2::~YsBoundingBoxMaker2()
{
	ClassInfo->Release();
}

void YsBoundingBoxMaker2::Begin(const YsVec2 &vec)
{
	min=vec;
	max=vec;
}

void YsBoundingBoxMaker2::Add(const YsVec2 &vec)
{
	min.Set
	   (YsSmaller(min.x(),vec.x()),
	    YsSmaller(min.y(),vec.y()));
	max.Set
	   (YsGreater(max.x(),vec.x()),
	    YsGreater(max.y(),vec.y()));
}

void YsBoundingBoxMaker2::Make(int n,const YsVec2 v[])
{
	if(n>0)
	{
		int i;
		Begin(v[0]);
		for(i=1; i<n; i++)
		{
			Add(v[i]);
		}
	}
}

void YsBoundingBoxMaker2::Make(YsList <YsVec2> *lst)
{
	if(lst!=NULL)
	{
		Begin(lst->dat);
	}
	while(lst!=NULL)
	{
		Add(lst->dat);
		lst=lst->Next();
	}
}

void YsBoundingBoxMaker2::Make(const YsArray <YsVec2> &ary)
{
	if(ary.GetNumItem()>0)
	{
		Begin(ary.GetItem(0));
	}
	int i;
	for(i=0; i<ary.GetNumItem(); i++)
	{
		Add(ary.GetItem(i));
	}
}

void YsBoundingBoxMaker2::Get(YsVec2 &n,YsVec2 &x) const
{
	n=min;
	x=max;
}

// Implementation //////////////////////////////////////////
YsClassHierarchy *YsBoundingBoxMaker3::ClassInfo=NULL;
char *YsBoundingBoxMaker3::ClassName="YsBoundingBoxMaker3";

YsBoundingBoxMaker3::YsBoundingBoxMaker3()
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsWizard::ClassInfo);
	}
	ClassInfo->Allocate();
}

YsBoundingBoxMaker3::YsBoundingBoxMaker3(const YsBoundingBoxMaker3 &from)
{
	if(ClassInfo==NULL)
	{
		ClassInfo=new YsClassHierarchy;
		ClassInfo->classname=WhatItIs();
		ClassInfo->SetIsChildOf(YsWizard::ClassInfo);
	}
	ClassInfo->Allocate();
	*this=from;
}

YsBoundingBoxMaker3::~YsBoundingBoxMaker3()
{
	ClassInfo->Release();
}

void YsBoundingBoxMaker3::Begin(const YsVec3 &vec)
{
	min=vec;
	max=vec;
}

void YsBoundingBoxMaker3::Add(const YsVec3 &vec)
{
	min.Set
	   (YsSmaller(min.x(),vec.x()),
	    YsSmaller(min.y(),vec.y()),
	    YsSmaller(min.z(),vec.z()));
	max.Set
	   (YsGreater(max.x(),vec.x()),
	    YsGreater(max.y(),vec.y()),
	    YsGreater(max.z(),vec.z()));
}

void YsBoundingBoxMaker3::Make(int n,const YsVec3 v[])
{
	if(n>0)
	{
		int i;
		Begin(v[0]);
		for(i=1; i<n; i++)
		{
			Add(v[i]);
		}
	}
}

void YsBoundingBoxMaker3::Make(const YsList <YsVec3> *lst)
{
	if(lst!=NULL)
	{
		Begin(lst->dat);
	}
	while(lst!=NULL)
	{
		Add(lst->dat);
		lst=lst->Next();
	}
}

void YsBoundingBoxMaker3::Make(const YsArray <YsVec3> &ary)
{
	if(ary.GetNumItem()>0)
	{
		Begin(ary.GetItem(0));
	}
	int i;
	for(i=0; i<ary.GetNumItem(); i++)
	{
		Add(ary.GetItem(i));
	}
}

void YsBoundingBoxMaker3::Get(YsVec3 &n,YsVec3 &x) const
{
	n=min;
	x=max;
}

